@@ -49,10 +49,10 @@
 
 (assert (probe-file (subpathname *output-directory* "cover-index.html")))
 
-;;; None of the code was executed
+;;; Only the top level forms have been executed
 (assert (zerop (sb-cover::ok-of (getf sb-cover::*counts* :branch))))
 (assert (zerop (sb-cover::all-of (getf sb-cover::*counts* :branch))))
-(assert (zerop (sb-cover::ok-of (getf sb-cover::*counts* :expression))))
+(assert (= 2 (sb-cover::ok-of (getf sb-cover::*counts* :expression))))
 (assert (plusp (sb-cover::all-of (getf sb-cover::*counts* :expression))))
 
 ;;; Call the function again
@@ -116,10 +116,46 @@
 (report)
 
 ;; Complete expression coverage
-(assert (= 13
+(assert (= 14
            (sb-cover::ok-of (getf sb-cover::*counts* :expression))
            (sb-cover::all-of (getf sb-cover::*counts* :expression))))
 
+;; Make sure we simulate (in-package) correctly.
+(sb-cover:clear-coverage)
+(compile-load "test-data-4")
+(test4)
+(report)
+
+;;; And now we should have complete expression coverage
+(assert (zerop (sb-cover::ok-of (getf sb-cover::*counts* :branch))))
+(assert (zerop (sb-cover::all-of (getf sb-cover::*counts* :branch))))
+(assert (plusp (sb-cover::ok-of (getf sb-cover::*counts* :expression))))
+(assert (= (sb-cover::ok-of (getf sb-cover::*counts* :expression))
+           (sb-cover::all-of (getf sb-cover::*counts* :expression))))
+
+;; Make sure we handle non-local exits from function calls correctly.
+(sb-cover:clear-coverage)
+(compile-load "test-data-5")
+(outer)
+(report)
+
+(assert (zerop (sb-cover::ok-of (getf sb-cover::*counts* :branch))))
+(assert (zerop (sb-cover::all-of (getf sb-cover::*counts* :branch))))
+(assert (= 12 (sb-cover::ok-of (getf sb-cover::*counts* :expression))))
+(assert (= 16 (sb-cover::all-of (getf sb-cover::*counts* :expression))))
+
+;; And then ensure that non-local exits from local calls are handled
+;; correctly as well.
+(sb-cover:clear-coverage)
+(compile-load "test-data-6")
+(nlx-from-flet)
+(report)
+
+(assert (zerop (sb-cover::ok-of (getf sb-cover::*counts* :branch))))
+(assert (zerop (sb-cover::all-of (getf sb-cover::*counts* :branch))))
+(assert (= 7 (sb-cover::ok-of (getf sb-cover::*counts* :expression))))
+(assert (= 11 (sb-cover::all-of (getf sb-cover::*counts* :expression))))
+
 ;; Clean up after the tests
 (map nil #'delete-file
      (directory (merge-pathnames #p"*.html" *output-directory*)))
