@@ -5,68 +5,89 @@
 
 (in-package "SB-VM")
 
-;;; It is arbitrary whether each of the next 4 routines is named ALLOC-something,
-;;; exporting an additional entry named CONS-something, versus being named CONS-something
-;;; and exporting ALLOC-something.  It just depends on which of those you would like
-;;; to have to set and clear the low bit to match ALLOC-DISPATCH.
-;;; Think of "cons" as meaning the generic sense of "consing", not as actually
-;;; allocating conses, because fixed-sized non-cons object allocation may enter
-;;; through the cons entry point. Cons cells *must* use that entry point.
-(define-assembly-routine (alloc->rnn (:export cons->rnn)) ()
-  (inst or :byte (ea 8 rsp-tn) 1)
-  CONS->RNN
-  (with-registers-preserved (c xmm)
+(macrolet ((do-fprs (operation regset)
+             (multiple-value-bind (mnemonic fpr-align)
+                 (ecase regset
+                   (:xmm (values 'movaps 16))
+                   (:ymm (values 'vmovaps 32)))
+               `(progn
+                  ,@(loop for regno below 16
+                       collect
+                         (ecase operation
+                           (push
+                            `(inst ,mnemonic (ea ,(+ 8 (* regno fpr-align)) rsp-tn)
+                                   (sb-x86-64-asm::get-fpr ,regset ,regno)))
+                           (pop
+                            `(inst ,mnemonic (sb-x86-64-asm::get-fpr ,regset ,regno)
+                                   (ea ,(+ 8 (* regno fpr-align)) rsp-tn)))))))))
+  ;; Caller will have allocated 512+64+256 bytes above the stack-pointer
+  ;; prior to the CALL. Use that as the save area.
+  (define-assembly-routine (save-ymm) ()
+    (inst push rax-tn)
+    (inst push rdx-tn)
+    (inst mov rax-tn 7)
+    (zeroize rdx-tn)
+    ;; Zero the header
+    (loop for i from (+ 512 24) by 8
+          repeat 8
+          do
+          (inst mov (ea i rsp-tn) rdx-tn))
+    (inst xsave (ea 24 rsp-tn))
+    (inst pop rdx-tn)
+    (inst pop rax-tn))
+  (define-assembly-routine (restore-ymm) ()
+    (inst push rax-tn)
+    (inst push rdx-tn)
+    (inst mov rax-tn 7)
+    (zeroize rdx-tn)
+    (inst xrstor (ea 24 rsp-tn))
+    (inst pop rdx-tn)
+    (inst pop rax-tn))
+  ;; As above, but only 256 bytes of the save area are needed, the rest goes to waste.
+  (define-assembly-routine (save-xmm (:export fpr-save)) ()
+    fpr-save ; KLUDGE: this is element 4 of the entry point vector
+    (do-fprs push :xmm))
+  (define-assembly-routine (restore-xmm (:export fpr-restore)) ()
+    fpr-restore ; KLUDGE: this is element 6 of the entry point vector
+    (do-fprs pop :xmm)))
+
+(define-assembly-routine (alloc-tramp) ()
+  (with-registers-preserved (c)
     (inst mov rdi-tn (ea 16 rbp-tn))
-    (inst call (make-fixup 'alloc-dispatch :assembly-routine))
+    (inst call (make-fixup "alloc" :foreign))
     (inst mov (ea 16 rbp-tn) rax-tn))) ; result onto stack
 
-#+avx2
-(define-assembly-routine (alloc->rnn.avx2 (:export cons->rnn.avx2)) ()
-  (inst or :byte (ea 8 rsp-tn) 1)
-  CONS->RNN.AVX2
-  (with-registers-preserved (c ymm)
+(define-assembly-routine (list-alloc-tramp) () ; CONS, ACONS, LIST, LIST*
+  (with-registers-preserved (c)
     (inst mov rdi-tn (ea 16 rbp-tn))
-    (inst call (make-fixup 'alloc-dispatch :assembly-routine))
+    (inst call (make-fixup "alloc_list" :foreign))
     (inst mov (ea 16 rbp-tn) rax-tn))) ; result onto stack
 
-(define-assembly-routine (alloc->r11 (:export cons->r11) (:return-style :none)) ()
-  (inst or :byte (ea 8 rsp-tn) 1)
-  CONS->R11
-  (with-registers-preserved (c xmm :except r11-tn)
-    (inst mov rdi-tn (ea 16 rbp-tn))
-    (inst call (make-fixup 'alloc-dispatch :assembly-routine))
-    (inst mov r11-tn rax-tn))
-  (inst ret 8)) ; pop argument
+(define-assembly-routine (listify-&rest (:return-style :none)) ()
+  (with-registers-preserved (c)
+    (inst mov rdi-tn (ea 16 rbp-tn)) ; 1st C call arg
+    (inst mov rsi-tn (ea 24 rbp-tn)) ; 2nd C call arg
+    (inst call (make-fixup "listify_rest_arg" :foreign))
+    (inst mov (ea 24 rbp-tn) rax-tn)) ; result
+  (inst ret 8)) ; pop one argument; the unpopped word now holds the result
 
-#+avx2
-(define-assembly-routine (alloc->r11.avx2 (:export cons->r11.avx2) (:return-style :none)) ()
-  (inst or :byte (ea 8 rsp-tn) 1)
-  CONS->R11.AVX2
-  (with-registers-preserved (c ymm :except r11-tn)
-    (inst mov rdi-tn (ea 16 rbp-tn))
-    (inst call (make-fixup 'alloc-dispatch :assembly-routine))
-    (inst mov r11-tn rax-tn))
-  (inst ret 8)) ; pop argument
-
-(define-assembly-routine (alloc-dispatch (:return-style :none)) ()
-  ;; If RDI has a 0 in the low bit, then we're allocating cons cells.
-  ;; A 1 bit signifies anything other than cons cells, and is equivalent
-  ;; to "could this object consume large-object pages" in gencgc.
-  (inst test :byte rdi-tn 1)
-  (inst jmp :z (make-fixup "alloc_list" :foreign))
-  (inst xor :byte rdi-tn 1) ; clear the bit
-  (inst jmp  (make-fixup "alloc" :foreign)))
+(define-assembly-routine (make-list (:return-style :none)) ()
+  (with-registers-preserved (c)
+    (inst mov rdi-tn (ea 16 rbp-tn)) ; 1st C call arg
+    (inst mov rsi-tn (ea 24 rbp-tn)) ; 2nd C call arg
+    (inst call (make-fixup "make_list" :foreign))
+    (inst mov (ea 24 rbp-tn) rax-tn)) ; result
+  (inst ret 8)) ; pop one argument; the unpopped word now holds the result
 
 ;;; These routines are for the deterministic consing profiler.
 ;;; The C support routine's argument is the return PC.
-;;; FIXME: we're missing routines that preserve YMM. I guess nobody cares.
 (define-assembly-routine (enable-alloc-counter) ()
-  (with-registers-preserved (c xmm)
+  (with-registers-preserved (c)
     (inst lea rdi-tn (ea 8 rbp-tn))
     (pseudo-atomic () (inst call (make-fixup "allocation_tracker_counted" :foreign)))))
 
 (define-assembly-routine (enable-sized-alloc-counter) ()
-  (with-registers-preserved (c xmm)
+  (with-registers-preserved (c)
     (inst lea rdi-tn (ea 8 rbp-tn))
     (pseudo-atomic () (inst call (make-fixup "allocation_tracker_sized" :foreign)))))
 
@@ -135,52 +156,7 @@
   (inst jmp (object-slot-ea rax-tn closure-fun-slot fun-pointer-lowtag)))
 
 (define-assembly-routine (ensure-symbol-hash (:return-style :raw)) ()
-  #+avx2
-  (progn
-    (test-cpu-feature cpu-has-ymm-registers)
-    (inst jmp :z call-preserving-xmm-only)
-    (with-registers-preserved (lisp ymm :except r11-tn)
-      (inst mov rdx-tn (ea 16 rbp-tn)) ; arg
-      (call-static-fun 'ensure-symbol-hash 1)
-      (inst mov (ea 16 rbp-tn) rdx-tn)) ; result to arg passing loc
-    (inst ret))
-  call-preserving-xmm-only
-  (with-registers-preserved (lisp xmm :except r11-tn)
+  (with-registers-preserved (lisp)
     (inst mov rdx-tn (ea 16 rbp-tn)) ; arg
     (call-static-fun 'ensure-symbol-hash 1)
     (inst mov (ea 16 rbp-tn) rdx-tn))) ; result to arg passing loc
-
-(define-assembly-routine (sb-impl::install-hash-table-lock
-                          (:return-style :raw))
-  ()
-  #+avx2
-  (progn
-    (test-cpu-feature cpu-has-ymm-registers)
-    (inst jmp :z call-preserving-xmm-only)
-    (with-registers-preserved (lisp ymm :except r11-tn)
-      (inst mov rdx-tn (ea 16 rbp-tn)) ; arg
-      (call-static-fun 'sb-impl::install-hash-table-lock 1)
-      (inst mov (ea 16 rbp-tn) rdx-tn)) ; result to arg passing loc
-    (inst ret))
-  call-preserving-xmm-only
-  (with-registers-preserved (lisp xmm :except r11-tn)
-    (inst mov rdx-tn (ea 16 rbp-tn)) ; arg
-    (call-static-fun 'sb-impl::install-hash-table-lock 1)
-    (inst mov (ea 16 rbp-tn) rdx-tn))) ; result to arg passing loc
-
-(define-assembly-routine (invalid-layout-trap (:return-style :none)) ()
-  #+avx2
-  (progn
-    (test-cpu-feature cpu-has-ymm-registers)
-    (inst jmp :z call-preserving-xmm-only)
-    (with-registers-preserved (lisp ymm :except r11-tn)
-      (inst mov rdx-tn (ea 16 rbp-tn)) ; arg
-      (call-static-fun 'update-object-layout 1)
-      (inst mov (ea 16 rbp-tn) rdx-tn)) ; result to arg passing loc
-    (inst ret)) ; Caller pops the result
-  call-preserving-xmm-only
-  (with-registers-preserved (lisp xmm :except r11-tn)
-    (inst mov rdx-tn (ea 16 rbp-tn)) ; arg
-    (call-static-fun 'update-object-layout 1)
-    (inst mov (ea 16 rbp-tn) rdx-tn)) ; result to arg passing loc
-  (inst ret)) ;  Caller pops the result
