@@ -9,7 +9,7 @@
 ;;;; bootstrap issues, SBCLification (e.g. DECLARE used to check
 ;;;; argument types), and other maintenance. Whether or not it then
 ;;;; supported all the environments implied by the reader conditionals
-;;;; in the source code (e.g. #!+CLOE-RUNTIME) before that
+;;;; in the source code (e.g. #+CLOE-RUNTIME) before that
 ;;;; modification, it sure doesn't now. It might perhaps, by blind
 ;;;; luck, be appropriate for some other CMU-CL-derived system, but
 ;;;; really it only attempts to be appropriate for SBCL.
@@ -65,7 +65,7 @@
 ;;;;      United States of America
 ;;;;      +1-617-221-1000
 
-(in-package "SB!LOOP")
+(in-package "SB-LOOP")
 
 ;;;; The design of this LOOP is intended to permit, using mostly the same
 ;;;; kernel of code, up to three different "loop" macros:
@@ -90,13 +90,15 @@
 
 ;;;; list collection macrology
 
-(sb!xc:defmacro with-loop-list-collection-head
+(sb-xc:defmacro with-loop-list-collection-head
     ((head-var tail-var &optional user-head-var) &body body)
   (let ((l (and user-head-var (list (list user-head-var nil)))))
     `(let* ((,head-var (list nil)) (,tail-var ,head-var) ,@l)
+       (declare (truly-dynamic-extent ,head-var)
+                ,@(and user-head-var `((list ,user-head-var))))
        ,@body)))
 
-(sb!xc:defmacro loop-collect-rplacd
+(sb-xc:defmacro loop-collect-rplacd
     (&environment env (head-var tail-var &optional user-head-var) form)
   (setq form (%macroexpand form env))
   (flet ((cdr-wrap (form n)
@@ -139,13 +141,13 @@
         (when user-head-var
           (setq answer
                 `(progn ,answer
-                        (setq ,user-head-var (cdr ,head-var)))))
+                        (setq ,user-head-var (sb-ext:truly-the list (cdr ,head-var))))))
         answer))))
 
-(sb!xc:defmacro loop-collect-answer (head-var
-                                                   &optional user-head-var)
-  (or user-head-var
-      `(cdr ,head-var)))
+(sb-xc:defmacro loop-collect-answer (head-var
+                                     &optional user-head-var)
+  `(sb-ext:truly-the list ,(or user-head-var
+                               `(cdr ,head-var))))
 
 ;;;; maximization technology
 
@@ -173,17 +175,18 @@ constructed.
   (flag-variable nil)
   (operations nil)
   (infinity-data nil :read-only t))
+(declaim (sb-ext:freeze-type loop-minimax))
 
 (defconstant-eqx +loop-minimax-type-infinities-alist+
-  ;; FIXME: Now that SBCL supports floating point infinities again, we
-  ;; should have floating point infinities here, as cmucl-2.4.8 did.
-  '((fixnum most-positive-fixnum most-negative-fixnum))
+    '((fixnum            most-positive-fixnum                  most-negative-fixnum)
+      (single-float      sb-ext:single-float-positive-infinity sb-ext:single-float-negative-infinity)
+      (double-float      sb-ext:double-float-positive-infinity sb-ext:double-float-negative-infinity))
   #'equal)
 
 (defun make-loop-minimax (answer-variable type)
   (let ((infinity-data (cdr (assoc type
                                    +loop-minimax-type-infinities-alist+
-                                   :test #'sb!xc:subtypep))))
+                                   :test #'subtypep))))
     (make-loop-minimax-internal
       :answer-variable answer-variable
       :type type
@@ -201,7 +204,7 @@ constructed.
           (gensym "LOOP-MAXMIN-FLAG-")))
   operation)
 
-(sb!xc:defmacro with-minimax-value (lm &body body)
+(sb-xc:defmacro with-minimax-value (lm &body body)
   (let ((init (loop-typed-init (loop-minimax-type lm)))
         (which (car (loop-minimax-operations lm)))
         (infinity-data (loop-minimax-infinity-data lm))
@@ -220,7 +223,7 @@ constructed.
            (declare (type ,type ,answer-var ,temp-var))
            ,@body))))
 
-(sb!xc:defmacro loop-accumulate-minimax-value (lm operation form)
+(sb-xc:defmacro loop-accumulate-minimax-value (lm operation form)
   (let* ((answer-var (loop-minimax-answer-variable lm))
          (temp-var (loop-minimax-temp-variable lm))
          (flag-var (loop-minimax-flag-variable lm))
@@ -265,7 +268,7 @@ code to be loaded.
   (and (symbolp loop-token)
        (values (gethash (symbol-name loop-token) table))))
 
-(sb!xc:defmacro loop-store-table-data (symbol table datum)
+(sb-xc:defmacro loop-store-table-data (symbol table datum)
   `(setf (gethash (symbol-name ,symbol) ,table) ,datum))
 
 ;;; "4.2.3 Type Specifiers" lists the standardized atomic type specifiers.
@@ -276,9 +279,7 @@ code to be loaded.
   ;; most atoms. The real test is whether type = :PRIMITIVE, as there are CL
   ;; symbols naming types which are not standard builtin type specifiers, e.g.
   ;; ARRAY-RANK, CHAR-CODE. (One almost wonders if that is technically wrong).
-  ;; At any rate, in the cross-compiler, we must not rely on the host's idea
-  ;; of SYMBOL-PACKAGE as it does not accurately model the target per se.
-  (and #-sb-xc-host (eq (symbol-package symbol) *cl-package*)
+  (and (eq (sb-xc:symbol-package symbol) *cl-package*)
        (or (eq (info :type :kind symbol) :primitive)
            ;; allow certain :instance types, but not all of them
            (member symbol '(hash-table package pathname random-state readtable)))
@@ -293,6 +294,7 @@ code to be loaded.
   iteration-keywords   ; hash table, value = (fn-name . extra-data)
   for-keywords         ; hash table, value = (fn-name . extra-data)
   path-keywords)       ; hash table, value = (fn-name . extra-data)
+(declaim (sb-ext:freeze-type loop-universe))
 (defmethod print-object ((u loop-universe) stream)
   (print-unreadable-object (u stream :type t :identity t)))
 
@@ -323,10 +325,10 @@ code to be loaded.
 (defun loop-make-desetq (var-val-pairs)
   (if var-val-pairs (cons 'loop-desetq var-val-pairs)))
 
-(sb!ext:defglobal *loop-desetq-temporary*
+(sb-ext:defglobal *loop-desetq-temporary*
         (make-symbol "LOOP-DESETQ-TEMP"))
 
-(sb!xc:defmacro loop-desetq (&environment env &rest var-val-pairs)
+(sb-xc:defmacro loop-desetq (&environment env &rest var-val-pairs)
   (labels ((find-non-null (var)
              ;; See whether there's any non-null thing here. Recurse
              ;; if the list element is itself a list.
@@ -415,35 +417,35 @@ code to be loaded.
 
 ;;; This is the pointer to the original, for things like NAMED that
 ;;; insist on being in a particular position
-  (original-source-code)
+  (original-source-code nil)
 
 ;;; This is (source-code *loop*) as of the "last" clause. It is used
 ;;; primarily for generating error messages (see loop-error, loop-warn).
-  (source-context)
+  (source-context nil)
 
 ;;; list of names for the LOOP, supplied by the NAMED clause
-  (names)
+  (names nil)
 
 ;;; The macroexpansion environment given to the macro.
   (macro-environment)
 
 ;;; This holds variable names specified with the USING clause.
 ;;; See LOOP-NAMED-VAR.
-  (named-vars)
+  (named-vars nil)
 
 ;;; LETlist-like list being accumulated for current group of bindings.
-  (vars)
+  (vars nil)
 
 ;;; List of declarations being accumulated in parallel with
 ;;; VARS.
-  (declarations)
+  (declarations nil)
 
 ;;; Declarations for destructuring bindings
-  (desetq-declarations)
+  (desetq-declarations nil)
 
 ;;; This is used by LOOP for destructuring binding, if it is doing
 ;;; that itself. See LOOP-MAKE-VAR.
-  (desetq)
+  (desetq nil)
 
 ;;; list of wrapping forms, innermost first, which go immediately
 ;;; inside the current set of parallel bindings being accumulated in
@@ -452,60 +454,61 @@ code to be loaded.
 ;;;   ((WITH-OPEN-FILE (G0001 G0002 ...))),
 ;;; with G0002 being one of the bindings in VARS (This is why the
 ;;; wrappers go inside of the variable bindings).
-  (wrappers)
+  (wrappers nil)
 
 ;;; This accumulates lists of previous values of VARS and the other
 ;;; lists above, for each new nesting of bindings. See
 ;;; LOOP-BIND-BLOCK.
-  (bind-stack)
+  (bind-stack nil)
 
 ;;; list of prologue forms of the loop, accumulated in reverse order
-  (prologue)
+  (prologue nil)
 
-  (before-loop)
-  (after-body)
+  (before-loop nil)
+  (after-body nil)
 
 ;;; This is T if we have emitted any body code, so that iteration
 ;;; driving clauses can be disallowed. This is not strictly the same
 ;;; as checking (BODY *LOOP*), because we permit some clauses such as
 ;;; RETURN to not be considered "real" body (so as to permit the user
 ;;; to "code" an abnormal return value "in loop").
-  (emitted-body)
+  (emitted-body nil)
 
 ;;; list of epilogue forms (supplied by FINALLY generally), accumulated
 ;;; in reverse order
-  (epilogue)
+  (epilogue nil)
 
 ;;; list of epilogue forms which are supplied after the above "user"
 ;;; epilogue. "Normal" termination return values are provide by
 ;;; putting the return form in here. Normally this is done using
 ;;; LOOP-EMIT-FINAL-VALUE, q.v.
-  (after-epilogue)
+  (after-epilogue nil)
 
 ;;; the "culprit" responsible for supplying a final value from the
 ;;; loop. This is so LOOP-DISALLOW-AGGREGATE-BOOLEANS can moan about
 ;;; disallowed anonymous collections.
-  (final-value-culprit)
+  (final-value-culprit nil)
 
 ;;; If not NIL, this is a temporary bound around the loop for holding
 ;;; the temporary value for "it" in things like "when (f) collect it".
 ;;; It may be used as a supertemporary by some other things.
-  (when-it-var)
+  (when-it-var nil)
 
 ;;; Sometimes we decide we need to fold together parts of the loop,
 ;;; but some part of the generated iteration code is different for the
 ;;; first and remaining iterations. This variable will be the
 ;;; temporary which is the flag used in the loop to tell whether we
 ;;; are in the first or remaining iterations.
-  (never-stepped-var)
+  (never-stepped-var nil)
 
 ;;; list of all the value-accumulation descriptor structures in the
 ;;; loop. See LOOP-GET-COLLECTION-INFO.
-  (collection-cruft) ; for multiple COLLECTs (etc.)
+  (collection-cruft nil) ; for multiple COLLECTs (etc.)
 
 ;;; This is the "current" loop context in use when we are expanding a
 ;;; loop. It gets bound on each invocation of LOOP.
   (universe nil :type loop-universe))
+(declaim (sb-ext:freeze-type macro-state))
 
 (defvar *loop*)
 (declaim (type macro-state *loop*))
@@ -519,10 +522,10 @@ code to be loaded.
 ;;;; code analysis stuff
 
 (defun loop-constant-fold-if-possible (form &optional expected-type)
-  (let* ((constantp (sb!xc:constantp form))
+  (let* ((constantp (constantp form))
          (value (and constantp (constant-form-value form))))
     (when (and constantp expected-type)
-      (unless (sb!xc:typep value expected-type)
+      (unless (sb-xc:typep value expected-type)
         (loop-warn "~@<The form ~S evaluated to ~S, which was not of ~
                     the anticipated type ~S.~:@>"
                    form value expected-type)
@@ -574,7 +577,7 @@ code to be loaded.
                              &optional (default-type required-type))
   (if (null specified-type)
       default-type
-      (multiple-value-bind (a b) (sb!xc:subtypep specified-type required-type)
+      (multiple-value-bind (a b) (subtypep specified-type required-type)
         (cond ((not b)
                (loop-warn "LOOP couldn't verify that ~S is a subtype of the required type ~S."
                           specified-type required-type))
@@ -607,7 +610,7 @@ code to be loaded.
                         (push (car cdr) result))))))
       (values (transform tree) ignores))))
 
-(sb!xc:defmacro loop-destructuring-bind
+(sb-xc:defmacro loop-destructuring-bind
     (lambda-list args &rest body)
   (multiple-value-bind (d-lambda-list ignores)
       (transform-destructuring lambda-list)
@@ -735,22 +738,22 @@ code to be loaded.
 
 (defun loop-typed-init (data-type &optional step-var-p)
   ;; FIXME: can't tell if unsupplied or NIL, but it has to be rare.
-  (when data-type
-   (let ((ctype (specifier-type data-type)))
-     ;; FIXME: use the ctype for the rest of the type operations, now
-     ;; that it's parsed.
-     (cond ((eql ctype *empty-type*)
-            (values nil t))
-           ((sb!xc:subtypep data-type 'number)
+  ;; FIXME: returns either 2 values or 1 value, which is poor style.
+  (unless data-type
+    (return-from loop-typed-init (values nil nil)))
+  (let ((ctype (specifier-type data-type)))
+    (when (eq ctype *empty-type*)
+      (return-from loop-typed-init (values nil t)))
+    (cond ((csubtypep ctype (specifier-type 'number))
             (let ((init (if step-var-p 1 0)))
               (flet ((like (&rest types)
-                       (coerce init (find-if (lambda (type)
-                                               (sb!xc:subtypep data-type type))
+                       (coerce init (find-if (lambda (spec)
+                                               (csubtypep ctype (specifier-type spec)))
                                              types))))
-                (cond ((sb!xc:subtypep data-type 'float)
+                (cond ((csubtypep ctype (specifier-type 'float))
                        (like 'single-float 'double-float
                              'short-float 'long-float 'float))
-                      ((sb!xc:subtypep data-type '(complex float))
+                      ((csubtypep ctype (specifier-type '(complex float)))
                        (like '(complex single-float)
                              '(complex double-float)
                              '(complex short-float)
@@ -758,18 +761,23 @@ code to be loaded.
                              '(complex float)))
                       (t
                        init)))))
-           ((sb!xc:subtypep data-type 'vector)
-            (when (array-type-p ctype)
-              (let ((etype (type-*-to-t
-                            (array-type-specialized-element-type ctype))))
-                (make-array 0 :element-type (type-specifier etype)))))
-           #!+sb-unicode
-           ((sb!xc:subtypep data-type 'extended-char)
+          ((csubtypep ctype (specifier-type 'vector))
+           (cond ((array-type-p ctype)
+                  (let ((etype (type-*-to-t
+                                (array-type-specialized-element-type ctype))))
+                    (make-array 0 :element-type (type-specifier etype))))
+                 ((csubtypep ctype (specifier-type 'string))
+                  "")))
+           #+sb-unicode
+           ((csubtypep ctype (specifier-type 'extended-char))
+            #+sb-xc-host
+            (error "Unimplemented on cross-compiler.")
+            #-sb-xc-host
             (code-char base-char-code-limit))
-           ((sb!xc:subtypep data-type 'character)
+           ((csubtypep ctype (specifier-type 'character))
             #\x)
            (t
-            nil)))))
+            nil))))
 
 (defun loop-optional-type (&optional variable &aux (loop *loop*))
   ;; No variable specified implies that no destructuring is permissible.
@@ -901,6 +909,12 @@ code to be loaded.
          (check-var-name name)
          (loop-declare-var name dtype :step-var-p step-var-p
                                       :initialization initialization)
+         ;; IGNORABLEize every variable because neither binding nor assignment constitutes
+         ;; a "use". Unfortunately there is no syntax for declaring what to ignore in LOOP.
+         ;; The idiom is to use NIL as a variable, but sometimes users don't, because they
+         ;; want variables names as information within a destructuring operation,
+         ;;  e.g. (for (this . that-not-used) in stuff do (frob this))
+         (push `(ignorable ,name) (declarations loop))
          ;; We use ASSOC on this list to check for duplications (above),
          ;; so don't optimize out this list:
          (push (list name (or initialization (loop-typed-init dtype step-var-p)))
@@ -932,21 +946,30 @@ code to be loaded.
             ;; longer be empty and the compiler won't produce
             ;; warnings.
             type)
-           ((sb!xc:typep init type)
+           ((sb-xc:typep init type)
             type)
-           ((sb!xc:typep init '(simple-array * (*)))
+           ((sb-xc:typep init '(simple-array * (*)))
+            ;; The cross-compiler must not inquire of the host
+            ;; for an array's element type.
+            #+sb-xc-host (bug "Can't get here")
             ;; type-of lets the size in
             `(or (simple-array ,(array-element-type init) (*)) ,type))
            (t
-            `(or ,(sb!xc:type-of init) ,type)))
+            (let ((disjunct
+                   #+sb-xc-host (if (null init)
+                                    'null
+                                    (error "Unexpected need for (TYPE-OF ~S) in xc"
+                                           init))
+                   #-sb-xc-host (type-of init)))
+              `(or ,disjunct ,type))))
      init)))
 
 (defun loop-declare-var (name dtype &key step-var-p initialization
                                          desetq &aux (loop *loop*))
   (cond ((or (null name) (null dtype) (eq dtype t)) nil)
         ((symbolp name)
-         (unless (or (sb!xc:subtypep t dtype)
-                     (and (eq (find-package :cl) (symbol-package name))
+         (unless (or (subtypep t dtype)
+                     (and (eq (sb-xc:symbol-package name) *cl-package*)
                           (eq :special (info :variable :kind name))))
            (let ((dtype `(type ,(if initialization
                                     dtype
@@ -1057,8 +1080,9 @@ code to be loaded.
   specified-type
   dtype
   (data nil)) ;collector-specific data
+(declaim (sb-ext:freeze-type loop-collector))
 
-(sb!xc:defmacro with-sum-count (lc &body body)
+(sb-xc:defmacro with-sum-count (lc &body body)
   (let* ((type (loop-collector-dtype lc))
          (temp-var (car (loop-collector-tempvars lc))))
     (multiple-value-bind (type init)
@@ -1344,7 +1368,7 @@ code to be loaded.
 
 (defun loop-for-across (var val data-type)
   (loop-make-var var nil data-type)
-  (let ((vector-var (gensym "LOOP-ACROSS-VECTOR-"))
+  (let ((vector-var (sb-xc:gensym "LOOP-ACROSS-VECTOR-"))
         (index-var (gensym "LOOP-ACROSS-INDEX-")))
     (multiple-value-bind (vector-form constantp vector-value)
         (loop-constant-fold-if-possible val 'vector)
@@ -1368,19 +1392,11 @@ code to be loaded.
 ;;;; list iteration
 
 (defun loop-list-step (listvar)
-  ;; We are not equipped to analyze whether 'FOO is the same as #'FOO
-  ;; here in any sensible fashion, so let's give an obnoxious warning
-  ;; whenever 'FOO is used as the stepping function.
-  ;;
-  ;; While a Discerning Compiler may deal intelligently with
-  ;; (FUNCALL 'FOO ...), not recognizing FOO may defeat some LOOP
-  ;; optimizations.
   (let ((stepper (cond ((loop-tequal (car (source-code *loop*)) :by)
                         (loop-pop-source)
                         (loop-get-form))
                        (t '(function cdr)))))
     (cond ((and (consp stepper) (eq (car stepper) 'quote))
-           (loop-warn "Use of QUOTE around stepping function in LOOP will be left verbatim.")
            `(funcall ,stepper ,listvar))
           ((and (consp stepper) (eq (car stepper) 'function))
            (list (cadr stepper) listvar))
@@ -1393,15 +1409,17 @@ code to be loaded.
       (loop-constant-fold-if-possible val)
     (let ((listvar var))
       (cond ((and var (symbolp var))
-             (loop-make-var var list data-type))
+             (loop-make-var var
+                            ;; Don't want to assert the type, as ENDP will do that
+                            `(the* (list :use-annotations t :source-form ,list) ,list)
+                            data-type))
             (t
              (loop-make-var (setq listvar (gensym)) list 't)
              (loop-make-var var nil data-type)))
       (let ((list-step (loop-list-step listvar)))
         (let* ((first-endtest
-                ;; mysterious comment from original CMU CL sources:
-                ;;   the following should use `atom' instead of `endp',
-                ;;   per [bug2428]
+                ;; the following should use `atom' instead of `endp',
+                ;; per 6.1.2.1.3
                 `(atom ,listvar))
                (other-endtest first-endtest))
           (when (and constantp (listp list-value))
@@ -1422,7 +1440,10 @@ code to be loaded.
       (loop-constant-fold-if-possible val)
     (let ((listvar (gensym "LOOP-LIST-")))
       (loop-make-var var nil data-type)
-      (loop-make-var listvar list t)
+      (loop-make-var listvar
+                     ;; Don't want to assert the type, as ENDP will do that
+                     `(the* (list :use-annotations t :source-form ,list) ,list)
+                     t)
       (let ((list-step (loop-list-step listvar)))
         (let* ((first-endtest `(endp ,listvar))
                (other-endtest first-endtest)
@@ -1444,6 +1465,7 @@ code to be loaded.
   (inclusive-permitted nil :read-only t)
   (function nil :read-only t)
   (user-data nil :read-only t))
+(declaim (sb-ext:freeze-type loop-path))
 
 (defun add-loop-path (names function universe
                       &key preposition-groups inclusive-permitted user-data)
@@ -1591,8 +1613,8 @@ code to be loaded.
          (limit-given nil) ; T when prep phrase has specified end
          (limit-constantp nil)
          (limit-value nil))
-     ;; Work around some "Unused" (rather, assigned-but-never-set) warnings
-     #+(or ccl clisp) (declare (ignorable start-constantp start-value))
+     ;; Silence the assigned-but-never-set warnings that CCL and CLISP emit
+     (declare (ignorable start-constantp start-value))
      (flet ((assert-index-for-arithmetic (index)
               (unless (atom index)
                 (loop-error "Arithmetic index must be an atom."))))
@@ -1628,14 +1650,14 @@ code to be loaded.
             (setq endform (if limit-constantp
                               `',limit-value
                               (loop-make-var
-                                 (gensym "LOOP-LIMIT-") form
+                                 (sb-xc:gensym "LOOP-LIMIT-") form
                                  `(and ,indexv-type real)))))
            (:by
             (multiple-value-setq (form stepby-constantp stepby)
               (loop-constant-fold-if-possible form
                                               `(and ,indexv-type (real (0)))))
             (unless stepby-constantp
-              (loop-make-var (setq stepby (gensym "LOOP-STEP-BY-"))
+              (loop-make-var (setq stepby (sb-xc:gensym "LOOP-STEP-BY-"))
                  form
                  `(and ,indexv-type (real (0)))
                  t)))
@@ -1831,7 +1853,7 @@ code to be loaded.
 
 ;;;; ANSI LOOP
 
-(sb!ext:define-load-time-global *loop-ansi-universe*
+(sb-ext:define-load-time-global *loop-ansi-universe*
   (let ((w (!make-standard-loop-universe
              :keywords '((named (loop-do-named))
                          (initially (loop-do-initially))
@@ -1923,12 +1945,12 @@ code to be loaded.
       (let ((tag (gensym)))
         `(block nil (tagbody ,tag (progn ,@keywords-and-forms) (go ,tag))))))
 
-(sb!xc:defmacro loop (&environment env &rest keywords-and-forms)
+(sb-xc:defmacro loop (&environment env &rest keywords-and-forms)
   (loop-standard-expansion keywords-and-forms
                            env
                            *loop-ansi-universe*))
 
-(sb!xc:defmacro loop-finish ()
+(sb-xc:defmacro loop-finish ()
   "Cause the iteration to terminate \"normally\", the same as implicit
 termination by an iteration driving clause, or by use of WHILE or
 UNTIL -- the epilogue code (if any) will be run, and any implicitly
