@@ -21,6 +21,10 @@
 ;;; and running the cross-compiler to produce target FASL files).
 (defpackage "SB-COLD" (:use "CL"))
 
+#+nil ; change to #+sbcl if desired, but at your own risk!
+(when (sb-sys:find-dynamic-foreign-symbol-address "show_gc_generation_throughput")
+  (setf (extern-alien "show_gc_generation_throughput" int) 1))
+
 (in-package "SB-COLD")
 
 (defun parse-make-host-parallelism (str)
@@ -164,24 +168,42 @@
     (rename-file x path)))
 (compile 'rename-file-a-la-unix)
 
-(export '(prepend-genfile-path read-from-file *generated-sources-root*))
+(export '(*target-sbcl-version* *generated-sources-root*
+          stem-source-path find-bootstrap-file read-from-file))
+(defvar *sources-root* "")
 (defvar *generated-sources-root* "")
+(defvar *src-cold-shared-pathname* *load-pathname*)
 
 ;;; See remark in COMPILE-STEM about strings vs. The Common Lisp Way
-(defun prepend-genfile-path (namestring)
-  (concatenate 'string
-               ;; if exact match to "output/", or mismatch at the next character
-               (if (member (mismatch "output/" namestring) '(nil 7))
-                   *generated-sources-root*
-                   "")
-               namestring))
-(compile 'prepend-genfile-path) ; seems in vogue to compile everything in this file
+(defun find-bootstrap-file (namestring)
+  (cond ((char= (char namestring 0) #\^)
+         ;; If it starts with a "^" then it means "src/cold/..."
+         (let ((this *src-cold-shared-pathname*)
+               (name (subseq namestring 1)))
+           (make-pathname :host (pathname-host this)
+                          :device (pathname-device this)
+                          :directory (pathname-directory this)
+                          :name (pathname-name name)
+                          :type (or (pathname-type name) (pathname-type this)))))
+        ((find #\/ namestring)
+         ;; Otherwise if it contains a slash, then it's a source file which is either
+         ;; in the tree as checked in, or generated by a prior build step.
+         (concatenate 'string
+                      (if (eql (mismatch "output/" namestring) 7) ; a generated source
+                          *generated-sources-root*
+                          *sources-root*)
+                      namestring))
+        (t
+         ;; Else, it's an optional user-supplied customization file,
+         ;; or a generated data file in the root directory such as "version.lisp-expr"
+         namestring)))
+(compile 'find-bootstrap-file) ; seems in vogue to compile everything in this file
 
 ;;; Return an expression read from the file named NAMESTRING.
 ;;; For user-supplied inputs, protect against more than one expression
-;;; appearing in the file. With trusted inputs we needn't bother.
+;;; appearing in the file. For in-tree inputs we needn't bother.
 (defun read-from-file (namestring &optional (enforce-single-expr t))
-  (with-open-file (s (prepend-genfile-path namestring))
+  (with-open-file (s (find-bootstrap-file namestring))
     (let* ((result (read s))
            (eof-result (cons nil nil)))
       (unless enforce-single-expr
@@ -202,7 +224,7 @@
 #+sbcl
 (progn
   (setq cl:*compile-print* nil)
-  (load "src/cold/muffler.lisp")
+  (load (find-bootstrap-file "^muffler"))
   ;; Let's just say we never care to see these.
   (declaim (sb-ext:muffle-conditions
             (satisfies unable-to-optimize-note-p)
@@ -212,16 +234,17 @@
 ;;;; special read-macros for building the cold system (and even for
 ;;;; building some of our tools for building the cold system)
 
-(load "src/cold/shebang.lisp")
+(load (find-bootstrap-file "^shebang"))
 
-(defvar *shebang-backend-subfeatures*
-  (let* ((default-subfeatures nil)
-         (customizer-file-name "customize-backend-subfeatures.lisp")
-         (customizer (if (probe-file customizer-file-name)
-                         (compile nil
-                                  (read-from-file customizer-file-name))
-                         #'identity)))
-    (funcall customizer default-subfeatures)))
+;;; Subfeatures could be assigned as late as the beginning of make-host-2,
+;;; but I don't want to introduce another mechanism for delaying reading
+;;; of the customizer just because we can.
+;;; But it's not well-advertised; does it really merit a customization file?
+(export 'backend-subfeatures)
+(defvar backend-subfeatures
+  (let ((customizer-file-name "customize-backend-subfeatures.lisp"))
+    (when (probe-file customizer-file-name)
+      (copy-list (funcall (compile nil (read-from-file customizer-file-name)) nil)))))
 
 ;;; When cross-compiling, the *FEATURES* set for the target Lisp is
 ;;; not in general the same as the *FEATURES* set for the host Lisp.
@@ -237,46 +260,49 @@
 ;;; The compromise is to examine a variable specifying a path
 ;;; (and it can't go in SB-COLD because the package is not made soon enough)
 (setf sb-xc:*features*
-      (let* ((pathname (let ((var 'cl-user::*sbcl-target-features-file*))
+      (let* ((pathname (let ((var 'cl-user::*sbcl-local-target-features-file*))
                          (if (boundp var)
                              (symbol-value var)
                              "local-target-features.lisp-expr")))
              (default-features
                (funcall (compile nil (read-from-file pathname))
-                        (read-from-file "base-target-features.lisp-expr")))
+                        (read-from-file "^base-target-features.lisp-expr")))
              (customizer-file-name "customize-target-features.lisp")
              (customizer (if (probe-file customizer-file-name)
                              (compile nil
                                       (read-from-file customizer-file-name))
                              #'identity))
-             (target-feature-list (funcall customizer default-features))
+             ;; Bind temporarily so that TARGET-FEATUREP and TARGET-PLATFORM-KEYWORD
+             ;; can see the tentative list.
+             (sb-xc:*features* (funcall customizer default-features))
              (gc (find-if (lambda (x) (member x '(:cheneygc :gencgc)))
-                          target-feature-list))
-             (arch (target-platform-keyword target-feature-list)))
+                          sb-xc:*features*))
+             (arch (target-platform-keyword)))
         ;; Win32 conditionally adds :sb-futex in grovel-features.sh
-        (when (featurep '(:and :sb-thread (:or :linux :freebsd)) target-feature-list)
-          (pushnew :sb-futex target-feature-list))
-        (when (featurep '(:and :sb-thread (:not :win32)) target-feature-list)
-          (push :pauseless-threadstart target-feature-list))
-        (when (featurep '(:and :sb-thread (:or :darwin :openbsd)) target-feature-list)
-          (push :os-thread-stack target-feature-list))
-        (when (and (member :x86 target-feature-list)
-                   (member :int4-breakpoints target-feature-list))
+        (when (target-featurep '(:and :sb-thread (:or :linux :freebsd)))
+          (pushnew :sb-futex sb-xc:*features*))
+        (when (target-featurep :immobile-space)
+          (pushnew :compact-instance-header sb-xc:*features*)
+          (pushnew :immobile-code sb-xc:*features*))
+        (when (target-featurep :64-bit)
+          (push :compact-symbol sb-xc:*features*))
+        (when (target-featurep '(:and :sb-thread (:not :win32)))
+          (push :pauseless-threadstart sb-xc:*features*))
+        (when (target-featurep '(:and :sb-thread (:or :darwin :openbsd)))
+          (push :os-thread-stack sb-xc:*features*))
+        (when (target-featurep '(:and :x86 :int4-breakpoints))
           ;; 0xCE is a perfectly good 32-bit instruction,
           ;; unlike on x86-64 where it is illegal. It's therefore
           ;; confusing to allow this feature in a 32-bit build.
           ;; But it's annoying to have a build script that otherwise works
           ;; for a native x86/x86-64 build except for needing one change.
           ;; Just print something and go on with life.
-          (setq target-feature-list
-                (remove :int4-breakpoints target-feature-list))
+          (setq sb-xc:*features* (remove :int4-breakpoints sb-xc:*features*))
           (warn "Removed :INT4-BREAKPOINTS from target features"))
-        (when (or (member :arm64 target-feature-list)
-                  (member :sse4 *shebang-backend-subfeatures*))
-          (push :round-float target-feature-list))
-        (when (and (member :arm64 target-feature-list)
-                  (member :darwin target-feature-list))
-          (push :arm-v8.1 *shebang-backend-subfeatures*))
+        (when (target-featurep '(:or :arm64 :sse4))
+          (push :round-float sb-xc:*features*))
+        (when (target-featurep '(:and :arm64 :darwin))
+          (push :arm-v8.1 backend-subfeatures))
 
         ;; Putting arch and gc choice first is visually convenient, versus
         ;; having to parse a random place in the line to figure out the value
@@ -284,7 +310,7 @@
         ;; De-duplicate the rest of the symbols because the command line
         ;; can add redundant --with-mumble options.
         (list* arch gc (sort (remove-duplicates
-                              (remove arch (remove gc target-feature-list)))
+                              (remove arch (remove gc sb-xc:*features*)))
                              #'string<))))
 
 ;;; Call for effect of signaling an error if no target picked.
@@ -307,12 +333,10 @@
          ("(and sb-safepoint (not sb-thread))" ":SB-SAFEPOINT requires :SB-THREAD")
          ("(and sb-thread (not (or riscv ppc ppc64 x86 x86-64 arm64)))"
           ":SB-THREAD not supported on selected architecture")
+         ("(and (not sb-thread) (or arm64 ppc64))"
+          "The selected architecture requires :SB-THREAD")
          ("(and gencgc cheneygc)"
           ":GENCGC and :CHENEYGC are incompatible")
-         ("(and cheneygc (not (or arm mips ppc riscv sparc)))"
-          ":CHENEYGC not supported on selected architecture")
-         ("(and gencgc (not (or sparc ppc ppc64 x86 x86-64 arm arm64 riscv)))"
-          ":GENCGC not supported on selected architecture")
          ("(not (or gencgc cheneygc))"
           "One of :GENCGC or :CHENEYGC must be enabled")
          ("(and sb-safepoint (not (and (or arm64 x86 x86-64) (or darwin linux win32))))"
@@ -320,9 +344,9 @@
          ("(not (or elf mach-o win32))"
           "No execute object file format feature defined")
          ("(and cons-profiling (not sb-thread))" ":CONS-PROFILING requires :SB-THREAD")
-         ("(and sb-linkable-runtime (not (or x86 x86-64)))"
+         ("(and sb-linkable-runtime (not (or arm arm64 x86 x86-64 ppc ppc64)))"
           ":SB-LINKABLE-RUNTIME not supported on selected architecture")
-         ("(and sb-linkable-runtime (not (or darwin linux win32)))"
+         ("(and sb-linkable-runtime (not (or darwin freebsd linux win32)))"
           ":SB-LINKABLE-RUNTIME not supported on selected operating system")
          ("(and sb-eval sb-fasteval)"
           ;; It sorta kinda works to have both, but there should be no need,
@@ -360,7 +384,7 @@
 ;;; All code depending on this is itself dependent on #+SB-SHOW.
 (defvar *cl-snapshot*)
 (when (member :sb-show sb-xc:*features*)
-  (load "src/cold/snapshot.lisp")
+  (load (find-bootstrap-file "^snapshot"))
   (setq *cl-snapshot* (take-snapshot "COMMON-LISP")))
 
 ;;;; master list of source files and their properties
@@ -396,14 +420,6 @@
     ;; :NOT-HOST is also set, since the SBCL assembler doesn't exist
     ;; while the cross-compiler is being built in the host ANSI Lisp.)
     :assem
-    ;; meaning: The #'COMPILE-STEM argument called :IGNORE-FAILURE-P
-    ;; should be true. (This is a KLUDGE: I'd like to get rid of it.
-    ;; For now, it exists so that compilation can proceed through the
-    ;; legacy warnings in src/compiler/x86/array.lisp, which I've
-    ;; never figured out but which were apparently acceptable in CMU
-    ;; CL. Eventually, it would be great to just get rid of all
-    ;; warnings and remove support for this flag. -- WHN 19990323)
-    :ignore-failure-p
     ;; meaning: ignore this flag.
     ;; This works around nonstandard behavior of "#." in certain hosts.
     ;; When the evaluated form yields 0 values, ECL and CLISP treat it
@@ -427,8 +443,8 @@
              (,flags (rest ,stem-and-flags)))
          ,@body))))
 
-;;; Given a STEM, remap the path component "/target/" to a suitable
-;;; target directory.
+;;; Given a STEM, remap the path components "/{arch}/" and "/asm-target/"
+;;; to suitable directories.
 (defun stem-remap-target (stem)
   (flet ((try-replacing (this that)
            (let ((position (search this stem)))
@@ -437,20 +453,20 @@
                             (subseq stem 0 (1+ position))
                             (string-downcase that)
                             (subseq stem (+ position (length this) -1)))))))
-    (or (try-replacing "/target/" (target-platform-keyword))
+    (or (try-replacing "/{arch}/" (target-platform-keyword))
         (try-replacing "/asm-target/" (backend-assembler-target-name))
         stem)))
 (compile 'stem-remap-target)
 
 ;;; Determine the source path for a stem by remapping from the abstract name
-;;; if it contains "/target/" and appending a ".lisp" suffix.
+;;; if it contains "/{arch}/" and appending a ".lisp" suffix.
 ;;; Assume that STEM is source-tree-relative unless it starts with "output/"
 ;;; in which case it could be elsewhere, if you prefer to keep the sources
 ;;; devoid of compilation artifacts. (The production of out-of-tree artifacts
 ;;; is not actually implemented in the generic build, however if your build
 ;;; system does that by itself, then hooray for you)
 (defun stem-source-path (stem)
-  (concatenate 'string (prepend-genfile-path (stem-remap-target stem)) ".lisp"))
+  (concatenate 'string (find-bootstrap-file (stem-remap-target stem)) ".lisp"))
 (compile 'stem-source-path)
 
 ;;; Determine the object path for a stem/flags/mode combination.
@@ -490,7 +506,7 @@
           ;; to produce warnings as a bug workaround.
           (let ((cl:*features* (cons feature cl:*features*))
                 (*readtable* *xc-readtable*))
-            (read-from-file "build-order.lisp-expr" nil))))
+            (read-from-file "^build-order.lisp-expr" nil))))
     (setf *stems-and-flags* (cons build-phase list)))
   ;; Now check for duplicate stems and bogus flags.
   (let ((stems (make-hash-table :test 'equal)))
@@ -529,42 +545,43 @@
 ;;; STEM and FLAGS are as per DO-STEMS-AND-FLAGS.  MODE is one of
 ;;; :HOST-COMPILE and :TARGET-COMPILE.
 (defun compile-stem (stem flags mode)
-
-  (let* (;; KLUDGE: Note that this CONCATENATE 'STRING stuff is not The Common
-         ;; Lisp Way, although it works just fine for common UNIX environments.
-         ;; Should it come to pass that the system is ported to environments
-         ;; where version numbers and so forth become an issue, it might become
-         ;; urgent to rewrite this using the fancy Common Lisp PATHNAME
-         ;; machinery instead of just using strings. In the absence of such a
-         ;; port, it might or might be a good idea to do the rewrite.
-         ;; -- WHN 19990815
-         (src (stem-source-path stem))
+  (let* ((src (stem-source-path stem))
          (obj (stem-object-path stem flags mode))
          ;; Compile-for-effect happens simultaneously with a forked compile,
          ;; so we need the for-effect output not to stomp on the real output.
          (tmp-obj
-          (concatenate 'string obj
-                       (if *compile-for-effect-only* "-scratch" "-tmp")))
-
+           (concatenate 'string obj
+                        (if *compile-for-effect-only* "-scratch" "-tmp")))
          (compile-file (ecase mode
                          (:host-compile
-                          #+ccl
+                          #+abcl ; ABCL complains about its own deficiency and then returns T
+                          ;; for warnings and failure. "Unable to compile function" is not our problem,
+                          ;; but I tried everything to muffle it, and nothing worked; so if it occurs,
+                          ;; treat the file as a success despite any actual problems that may exist.
+                          (lambda (&rest args)
+                            (let (compiler-bug)
+                              ;; Even though COMPILER-UNSUPPORTED-FEATURE-ERROR is a condition class,
+                              ;; HANDLER-BIND seems unable to match it. What the hell? Bugs all the way down.
+                              (handler-bind ((condition
+                                              (lambda (c)
+                                                (when (search "Using interpreted form" (princ-to-string c))
+                                                  (setq compiler-bug t)))))
+                                (multiple-value-bind (fasl warn err) (apply #'compile-file args)
+                                  (if compiler-bug (values fasl nil nil) (values fasl warn err))))))
+                          #+ccl ; CCL doesn't like NOTINLINE on unknown functions
                           (lambda (&rest args)
-                            ;; CCL doesn't like NOTINLINE on uknown functions
                             (handler-bind ((ccl:compiler-warning
                                              (lambda (c)
                                                (when (eq (ccl::compiler-warning-warning-type c)
                                                          :unknown-declaration-function)
                                                  (muffle-warning c)))))
                               (apply #'compile-file args)))
-                          #-ccl
-                          #'compile-file)
+                          #-(or abcl ccl) #'compile-file)
                          (:target-compile (if (find :assem flags)
                                               *target-assemble-file*
                                               *target-compile-file*))))
          (trace-file (if (find :trace-file flags) t nil))
-         (block-compile (if (find :block-compile flags) t :specified))
-         (ignore-failure-p (find :ignore-failure-p flags)))
+         (block-compile (if (find :block-compile flags) t :specified)))
     (declare (type function compile-file))
 
     (ensure-directories-exist obj :verbose cl:*compile-print*) ; host's value
@@ -615,26 +632,27 @@
        retry-compile-file
          (multiple-value-bind (output-truename warnings-p failure-p)
              (restart-case
-                 (apply compile-file src :output-file tmp-obj
-                          :block-compile (and ;; Block compilation was
-                                              ;; completely broken
-                                              ;; from the beginning of
-                                              ;; SBCL history until
-                                              ;; version 2.0.2.
-                                              #+sbcl
-                                              (or (eq mode :target-compile)
-                                                  (and (find-symbol "SPLIT-VERSION-STRING" "SB-C")
-                                                       (funcall (find-symbol "VERSION>=" "SB-C")
-                                                                (funcall (find-symbol "SPLIT-VERSION-STRING" "SB-C")
-                                                                         (lisp-implementation-version))
-                                                                '(2 0 2))))
-                                              block-compile)
-                          :allow-other-keys t
-                          ;; If tracing, also print, but don't specify :PRINT unless specifying
-                          ;; :TRACE-FILE so that whatever the default is for *COMPILE-PRINT*
-                          ;; prevails, insensitively to whether it's the SB-XC: or CL: symbol.
-                          (when trace-file
-                            '(:trace-file t :print t)))
+                 (apply compile-file src
+                        :output-file tmp-obj
+                        :block-compile (and
+                                        ;; Block compilation was
+                                        ;; completely broken from the
+                                        ;; beginning of SBCL history
+                                        ;; until version 2.0.2.
+                                        #+sbcl
+                                        (or (eq mode :target-compile)
+                                            (and (find-symbol "SPLIT-VERSION-STRING" "HOST-SB-C")
+                                                 (funcall (find-symbol "VERSION>=" "HOST-SB-C")
+                                                          (funcall (find-symbol "SPLIT-VERSION-STRING" "HOST-SB-C")
+                                                                   (lisp-implementation-version))
+                                                          '(2 0 2))))
+                                        block-compile)
+                        :allow-other-keys t
+                        ;; If tracing, also print, but don't specify :PRINT unless specifying
+                        ;; :TRACE-FILE so that whatever the default is for *COMPILE-PRINT*
+                        ;; prevails, insensitively to whether it's the SB-XC: or CL: symbol.
+                        (when trace-file
+                          '(:trace-file t :print t)))
                (recompile ()
                  :report report-recompile-restart
                  (go retry-compile-file)))
@@ -642,23 +660,20 @@
            (cond ((not output-truename)
                   (error "couldn't compile ~S" src))
                  (failure-p
-                  (if ignore-failure-p
-                      (warn "ignoring FAILURE-P return value from compilation of ~S"
-                            src)
-                      (unwind-protect
-                           (restart-case
-                               (error "FAILURE-P was set when creating ~S."
-                                      obj)
-                             (recompile ()
-                               :report report-recompile-restart
-                               (go retry-compile-file))
-                             (continue ()
-                               :report report-continue-restart
-                               (setf failure-p nil)))
-                        ;; Don't leave failed object files lying around.
-                        (when (and failure-p (probe-file tmp-obj))
-                          (delete-file tmp-obj)
-                          (format t "~&deleted ~S~%" tmp-obj)))))
+                  (unwind-protect
+                       (restart-case
+                           (error "FAILURE-P was set when creating ~S."
+                                  obj)
+                         (recompile ()
+                           :report report-recompile-restart
+                           (go retry-compile-file))
+                         (continue ()
+                           :report report-continue-restart
+                           (setf failure-p nil)))
+                    ;; Don't leave failed object files lying around.
+                    (when (and failure-p (probe-file tmp-obj))
+                      (delete-file tmp-obj)
+                      (format t "~&deleted ~S~%" tmp-obj))))
                  ;; Otherwise: success, just fall through.
                  (t nil)))))
 
@@ -667,7 +682,7 @@
     (cond ((not *compile-for-effect-only*)
            (rename-file-a-la-unix tmp-obj obj))
           ((probe-file tmp-obj)
-           (delete-file tmp-obj))) ; clean up the trash
+           (delete-file tmp-obj)))      ; clean up the trash
 
     ;; nice friendly traditional return value
     (pathname obj)))
@@ -676,6 +691,7 @@
 (defparameter *host-quirks*
   (or #+cmu  '(:host-quirks-cmu)
       #+ecl  '(:host-quirks-ecl)
+      #+ccl  '(:host-quirks-ccl)
       #+sbcl '(:host-quirks-sbcl))) ; not so much a "quirk", but consistent anyway
 
 ;;; Execute function FN in an environment appropriate for compiling the
