@@ -63,15 +63,13 @@
   (:args (x :scs (double-reg) :to :save))
   (:results (y))
   (:note "float to pointer coercion")
-  (:temporary (:sc non-descriptor-reg) flag)
-  (:temporary (:scs (interior-reg)) lip)
+  (:temporary (:scs (non-descriptor-reg) :offset lr-offset) lr)
   (:results (y :scs (descriptor-reg)))
   (:generator 13
-   (with-fixed-allocation (y flag
-                           double-float-widetag
-                           double-float-value-slot
-                           :lip lip)
-     (storew x y double-float-value-slot other-pointer-lowtag))))
+    (with-fixed-allocation (y lr
+                            double-float-widetag
+                            double-float-value-slot)
+      (storew x y double-float-value-slot other-pointer-lowtag))))
 
 (define-move-vop move-from-double :move (double-reg) (descriptor-reg))
 
@@ -144,7 +142,7 @@
     (:results (y :scs (complex-single-reg) :load-if (not (location= x y))))
     (:note "complex single float move")
     (:generator 0
-      (inst s-mov y x)))
+      (move-complex-double y x)))
 (define-move-vop complex-single-move :move
   (complex-single-reg) (complex-single-reg))
 
@@ -154,7 +152,7 @@
   (:results (y :scs (complex-double-reg) :load-if (not (location= x y))))
   (:note "complex double float move")
   (:generator 0
-    (inst s-mov y x)))
+    (move-complex-double y x)))
 (define-move-vop complex-double-move :move
   (complex-double-reg) (complex-double-reg))
 
@@ -166,30 +164,28 @@
 (define-vop (move-from-complex-single)
   (:args (x :scs (complex-single-reg) :to :save))
   (:results (y :scs (descriptor-reg)))
-  (:temporary (:sc non-descriptor-reg) pa-flag)
-  (:temporary (:scs (interior-reg)) lip)
+  (:temporary (:scs (non-descriptor-reg) :offset lr-offset) lr)
   (:note "complex single float to pointer coercion")
   (:generator 13
-     (with-fixed-allocation (y pa-flag complex-single-float-widetag
-                             complex-single-float-size :lip lip)
-       (inst str x (@ y (- (* complex-single-float-data-slot
-                              n-word-bytes)
-                           other-pointer-lowtag))))))
+    (with-fixed-allocation (y lr complex-single-float-widetag
+                            complex-single-float-size)
+      (inst str x (@ y (- (* complex-single-float-data-slot
+                             n-word-bytes)
+                          other-pointer-lowtag))))))
 (define-move-vop move-from-complex-single :move
   (complex-single-reg) (descriptor-reg))
 
 (define-vop (move-from-complex-double)
   (:args (x :scs (complex-double-reg) :to :save))
   (:results (y :scs (descriptor-reg)))
-  (:temporary (:sc non-descriptor-reg) pa-flag)
-  (:temporary (:scs (interior-reg)) lip)
+  (:temporary (:scs (non-descriptor-reg) :offset lr-offset) lr)
   (:note "complex double float to pointer coercion")
   (:generator 13
-     (with-fixed-allocation (y pa-flag complex-double-float-widetag
-                             complex-double-float-size :lip lip)
-       (inst str x (@ y (- (* complex-double-float-real-slot
-                              n-word-bytes)
-                           other-pointer-lowtag))))))
+    (with-fixed-allocation (y lr complex-double-float-widetag
+                            complex-double-float-size)
+      (inst str x (@ y (- (* complex-double-float-real-slot
+                             n-word-bytes)
+                          other-pointer-lowtag))))))
 (define-move-vop move-from-complex-double :move
   (complex-double-reg) (descriptor-reg))
 
@@ -243,7 +239,7 @@
   (:generator 2
     (sc-case y
       (complex-double-reg
-       (inst s-mov y x))
+       (move-complex-double y x))
       (complex-double-stack
        (storew x nfp (tn-offset y))))))
 (define-move-vop move-complex-double-float-arg :move-arg
@@ -802,3 +798,33 @@
   (:translate imagpart)
   (:note "complex double float imagpart")
   (:variant :imag))
+
+(define-vop ()
+  (:translate round-double)
+  (:policy :fast-safe)
+  (:args (x :scs (double-reg) :target r))
+  (:arg-types double-float (:constant symbol))
+  (:info mode)
+  (:results (r :scs (double-reg)))
+  (:result-types double-float)
+  (:generator 2
+    (ecase mode
+      (:round (inst frintn r x))
+      (:floor (inst frintm r x))
+      (:ceiling (inst frintp r x))
+      (:truncate (inst frintz r x)))))
+
+(define-vop ()
+  (:translate round-single)
+  (:policy :fast-safe)
+  (:args (x :scs (single-reg) :target r))
+  (:arg-types single-float (:constant symbol))
+  (:info mode)
+  (:results (r :scs (single-reg)))
+  (:result-types single-float)
+  (:generator 2
+    (ecase mode
+      (:round (inst frintn r x))
+      (:floor (inst frintm r x))
+      (:ceiling (inst frintp r x))
+      (:truncate (inst frintz r x)))))
