@@ -17,11 +17,26 @@
 ;;;; utilities used during code generation
 
 ;;; the number of bytes used by the code object header
-(defun component-header-length (&optional
-                                (component *component-being-compiled*))
-  (let* ((2comp (component-info component))
-         (constants (ir2-component-constants 2comp)))
-    (ash (align-up (length constants) code-boxed-words-align) sb-vm:word-shift)))
+
+(macrolet ((header-length-in-bytes (comp)
+             `(let* ((2comp (component-info ,comp))
+                     (constants (ir2-component-constants 2comp)))
+                (ash (align-up (length constants) code-boxed-words-align)
+                     sb-vm:word-shift))))
+  #-sb-xc-host
+  (defun component-header-length (&optional (component *component-being-compiled*))
+    (header-length-in-bytes component))
+  ;; KLUDGE: the assembler can not emit backpatches comprising jump tables without
+  ;; knowing the boxed code header length. But there is no compiler IR2 metaobject,
+  ;; for SB-FASL:*ASSEMBLER-ROUTINES*. We have to return a fixed answer for that.
+  #+sb-xc-host
+  (defun component-header-length ()
+    (if (boundp '*component-being-compiled*)
+        (header-length-in-bytes *component-being-compiled*)
+        (* sb-vm:n-word-bytes (align-up sb-vm:code-constants-offset 2)))))
+
+(defun component-n-jump-table-entries (&optional (component *component-being-compiled*))
+  (ir2-component-n-jump-table-entries (component-info component)))
 
 ;;; the size of the NAME'd SB in the currently compiled component.
 ;;; This is useful mainly for finding the size for allocating stack
@@ -34,9 +49,9 @@
 (defun current-nfp-tn (vop)
   (unless (zerop (sb-allocated-size 'non-descriptor-stack))
     (let ((block (ir2-block-block (vop-block vop))))
-    (when (ir2-physenv-number-stack-p
-           (physenv-info
-            (block-physenv block)))
+    (when (ir2-environment-number-stack-p
+           (environment-info
+            (block-environment block)))
       (ir2-component-nfp (component-info (block-component block)))))))
 
 ;;; the TN that is used to hold the number stack frame-pointer in the
@@ -44,13 +59,13 @@
 ;;; allocated
 (defun callee-nfp-tn (2env)
   (unless (zerop (sb-allocated-size 'non-descriptor-stack))
-    (when (ir2-physenv-number-stack-p 2env)
+    (when (ir2-environment-number-stack-p 2env)
       (ir2-component-nfp (component-info *component-being-compiled*)))))
 
 ;;; the TN used for passing the return PC in a local call to the function
 ;;; designated by 2ENV
 (defun callee-return-pc-tn (2env)
-  (ir2-physenv-return-pc-pass 2env))
+  (ir2-environment-return-pc-pass 2env))
 
 ;;;; noise to emit an instruction trace
 
@@ -86,9 +101,6 @@
            (block-home-lambda
             (block-next (component-head *component-being-compiled*))))
           (or (> speed compilation-speed) (> space compilation-speed))))
-(defun default-segment-inst-hook ()
-  (and *compiler-trace-output*
-       #'trace-instruction))
 
 ;;; Some platforms support unboxed constants immediately following the boxed
 ;;; code header. Such platform must implement supporting 4 functions:
@@ -114,22 +126,59 @@
 ;;; to get handles (as returned by sb-vm:inline-constant-value) from constant
 ;;; descriptors.
 ;;;
-#+(or x86 x86-64 arm64)
+#+(or arm64 mips ppc ppc64 x86 x86-64)
 (defun register-inline-constant (&rest constant-descriptor)
-  (declare (dynamic-extent constant-descriptor))
+  ;; N.B.: Please do not think yourself so clever as to declare DYNAMIC-EXTENT on
+  ;; CONSTANT-DESCRIPTOR. Giving the list indefinite extent allows backends to simply
+  ;; return it as the key to the hash-table in the most trivial possible implementation.
+  ;; The cost of listifying a &REST arg is unnoticeable here.
   (let ((asmstream *asmstream*)
         (constant (sb-vm:canonicalize-inline-constant constant-descriptor)))
-    (ensure-gethash
-     constant
-     (asmstream-constant-table asmstream)
-     (multiple-value-bind (label value) (sb-vm:inline-constant-value constant)
-       (vector-push-extend (cons constant label)
-                           (asmstream-constant-vector asmstream))
-       value))))
-#-(or x86 x86-64 arm64)
+    (values ; kill the second value
+     (ensure-gethash
+      constant
+      (asmstream-constant-table asmstream)
+      (multiple-value-bind (label value) (sb-vm:inline-constant-value constant)
+        (vector-push-extend (cons constant label)
+                            (asmstream-constant-vector asmstream))
+        value)))))
+#-(or arm64 mips ppc ppc64 x86 x86-64)
 (progn (defun sb-vm:sort-inline-constants (constants) constants)
        (defun sb-vm:emit-inline-constant (&rest args)
          (error "EMIT-INLINE-CONSTANT called with ~S" args)))
+;;; Emit the subset of inline constants which represent jump tables
+;;; and remove those constants so that they don't get emitted again.
+(defun emit-jump-tables (ir2-component)
+  ;; Other backends will probably need relative jump tables instead
+  ;; of absolute tables because of the problem of needing to load
+  ;; the LIP register prior to loading an arbitrary PC.
+  (let* ((asmstream *asmstream*)
+         (constants (asmstream-constant-vector asmstream))
+         (section (asmstream-data-section asmstream))
+         (nwords 0))
+    (collect ((jump-tables) (other))
+      (dovector (constant constants)
+        ;; Constant is ((category . data) . label)
+        (cond ((eq (caar constant) :jump-table)
+               (incf nwords (length (cdar constant)))
+               (jump-tables constant))
+              (t
+               (other constant))))
+      ;; Preface the unboxed data with a count of jump-table words,
+      ;; including the count itself as 1 word.
+      ;; On average this will add another padding word half of the time
+      ;; depending on the number of boxed constants. We could reduce space
+      ;; by storing one bit in the header indicating whether or not there
+      ;; is a jump table. I don't think that's worth the trouble.
+      (emit section `(.lispword ,(1+ nwords)))
+      (when (plusp nwords)
+        (setf (ir2-component-n-jump-table-entries ir2-component) nwords)
+        (dolist (constant (jump-tables))
+          (sb-vm:emit-inline-constant section (car constant) (cdr constant)))
+        (let ((nremaining (length (other))))
+          (adjust-array constants nremaining
+                        :fill-pointer nremaining
+                        :initial-contents (other)))))))
 ;;; Return T if and only if there were any constants emitted.
 (defun emit-inline-constants ()
   (let* ((asmstream *asmstream*)
@@ -219,13 +268,18 @@
 ;; Collect "static" count of number of times each vop is employed.
 ;; (as opposed to "dynamic" - how many times its code is hit at runtime)
 (defglobal *static-vop-usage-counts* nil)
+(defparameter *do-instcombine-pass* t)
 
-(defun generate-code (component)
+(defun generate-code (component &aux (ir2-component (component-info component)))
+  (declare (type ir2-component ir2-component))
   (when *compiler-trace-output*
-    (format *compiler-trace-output*
-            "~|~%assembly code for ~S~2%"
-            component))
+    (let ((*print-pretty* nil)) ; force 1 line
+      (format *compiler-trace-output* "~|~%assembly code for ~S~2%" component)))
   (let* ((prev-env nil)
+         (sb-vm::*adjustable-vectors* nil)
+         ;; The first function's alignment word is zero-filled, but subsequent
+         ;; ones can use a NOP which helps the disassembler not lose sync.
+         (filler-pattern 0)
          (asmstream (make-asmstream))
          (*asmstream* asmstream))
 
@@ -242,21 +296,24 @@
             ;; on 8 byte boundaries we cannot guarantee proper loop alignment
             ;; there (yet.)  Only x86-64 does something with ALIGNP, but
             ;; it may be useful in the future.
+            ;; FIXME: see comment in ASSEMBLE-SECTIONS - we *can* enforce larger
+            ;; alignment than the size of a cons cell.
             (let ((alignp (let ((cloop (block-loop 1block)))
                             (when (and cloop
                                        (loop-tail cloop)
                                        (not (loop-info cloop)))
                               ;; Mark the loop as aligned by saving the IR1 block aligned.
                               (setf (loop-info cloop) 1block)
-                              t))))
+                              filler-pattern))))
+              (setf filler-pattern :long-nop)
               (emit-block-header (block-label 1block)
                                  (ir2-block-%trampoline-label block)
                                  (ir2-block-dropped-thru-to block)
                                  alignp)))
-          (let ((env (block-physenv 1block)))
+          (let ((env (block-environment 1block)))
             (unless (eq env prev-env)
-              (let ((lab (gen-label)))
-                (setf (ir2-physenv-elsewhere-start (physenv-info env))
+              (let ((lab (gen-label "environment elsewhere start")))
+                (setf (ir2-environment-elsewhere-start (environment-info env))
                       lab)
                 (emit (asmstream-elsewhere-section asmstream) lab))
               (setq prev-env env)))))
@@ -264,7 +321,7 @@
           ((null vop))
         (let ((gen (vop-info-generator-function (vop-info vop))))
           (awhen *static-vop-usage-counts*
-            (let ((name (vop-info-name (vop-info vop))))
+            (let ((name (vop-name vop)))
               (incf (gethash name it 0))))
           (assemble (:code vop)
             (cond ((not gen)
@@ -281,29 +338,49 @@
                   (t
                    (funcall gen vop)))))))
 
-    ;; Truncate the final assembly code buffer to length
-    (sb-assem::truncate-section-to-length (asmstream-code-section asmstream))
+    (when *do-instcombine-pass*
+      #+(or arm64 x86-64)
+      (sb-assem::combine-instructions (asmstream-code-section asmstream)))
 
-    (coverage-mark-lowering-pass component asmstream)
+    (emit (asmstream-data-section asmstream)
+          (sb-assem::asmstream-data-origin-label asmstream))
+    ;; Jump tables precede the coverage mark bytes to simplify locating
+    ;; them in trans_code().
+    (emit-jump-tables ir2-component)
+    (let ((coverage-map (ir2-component-coverage-map ir2-component)))
+      (unless (zerop (length coverage-map))
+        ;; Nothing depends on the length of the constant vector at this
+        ;; phase (codegen has not made use of component-header-length),
+        ;; so extending can be done with impunity.
+        #+arm64
+        (vector-push-extend (cons :coverage-marks (length coverage-map))
+                            (ir2-component-constants ir2-component))
+        (vector-push-extend
+         (make-constant (cons 'coverage-map
+                              (map-into coverage-map #'car coverage-map)))
+         (ir2-component-constants ir2-component))
+        ;; Allocate space in the data section for coverage marks.
+        #-arm64
+        (emit (asmstream-data-section asmstream)
+              `(.skip ,(length coverage-map) #-(or x86 x86-64) #xff))))
 
     (emit-inline-constants)
-    (let* ((info (component-info component))
-           (simple-fun-labels
-            (mapcar #'entry-info-offset (ir2-component-entries info)))
-           (n-boxed (length (ir2-component-constants info)))
+
+    (let* ((n-boxed (length (ir2-component-constants ir2-component)))
            ;; Skew is either 0 or N-WORD-BYTES depending on whether the boxed
            ;; header length is even or odd
            (skew (if (and (= code-boxed-words-align 1) (oddp n-boxed))
                      sb-vm:n-word-bytes
                      0)))
       (multiple-value-bind (segment text-length fixup-notes fun-table)
-          (assemble-sections asmstream
-                             simple-fun-labels
-                             (make-segment :header-skew skew
-                                           :run-scheduler (default-segment-run-scheduler)
-                                           :inst-hook (default-segment-inst-hook)))
+          (assemble-sections
+           asmstream
+           (mapcar #'entry-info-offset (ir2-component-entries ir2-component))
+           (make-segment :header-skew skew
+                         :run-scheduler (default-segment-run-scheduler)))
         (values segment text-length fun-table
-                (asmstream-elsewhere-label asmstream) fixup-notes)))))
+                (asmstream-elsewhere-label asmstream) fixup-notes
+                (sb-assem::get-allocation-points asmstream))))))
 
 (defun label-elsewhere-p (label-or-posn kind)
   (let ((elsewhere (label-position *elsewhere-label*))
@@ -318,58 +395,3 @@
         ;; We're interested in what precedes the return, not after
         (< elsewhere label)
         (<= elsewhere label))))
-
-;;; Translate .COVERAGE-MARK pseudo-op into machine assembly language,
-;;; combining any number of consecutive operations with no intervening
-;;; control flow into a single operation.
-;;; FIXME: this pass runs even if no coverage instrumentation was generated.
-(defun coverage-mark-lowering-pass (component asmstream)
-  (declare (ignorable component asmstream))
-  #+(or x86-64 x86)
-  (let ((label (gen-label))
-        ;; vector of lists of original source paths covered
-        (src-paths (make-array 10 :fill-pointer 0))
-        (previous-mark))
-    (dolist (buffer (reverse (sb-assem::section-buf-chain
-                              (asmstream-code-section asmstream))))
-      (dotimes (i (length buffer))
-        (let ((item (svref buffer i)))
-          (typecase item
-           (label
-            (when (label-usedp item) ; control can transfer to here
-              (setq previous-mark nil)))
-           (function ; this can do anything, who knows
-            (setq previous-mark nil))
-           (t
-            (let ((mnemonic (first item)))
-              (when (vop-p mnemonic)
-                (pop item)
-                (setq mnemonic (first item)))
-              (cond ((branch-opcode-p mnemonic) ; control flow kills mark combining
-                     (setq previous-mark nil))
-                    ((eq mnemonic '.coverage-mark)
-                     (let ((path (second item)))
-                       (cond ((not previous-mark) ; record a new path
-                              (let ((mark-index
-                                     (vector-push-extend (list path) src-paths)))
-                                ;; have the backend lower it into a real instruction
-                                (replace-coverage-instruction buffer i label mark-index))
-                              (setq previous-mark t))
-                             (t ; record that the already-emitted mark pertains
-                                ; to an additional source path
-                              (push path (elt src-paths (1- (fill-pointer src-paths))))
-                              ;; turn this line into a (virtual) no-op
-                              (rplaca item '.comment))))))))))))
-    ;; Allocate space in the data section for coverage marks
-    (let ((mark-index (length src-paths)))
-      (when (plusp mark-index)
-        (setf (label-usedp label) t)
-        (let ((v (ir2-component-constants (component-info component))))
-          ;; Nothing depends on the length of the constant vector at this
-          ;; phase (codegen has not made use of component-header-length),
-          ;; so extending can be done with impunity.
-          (vector-push-extend
-           (make-constant (cons 'coverage-map
-                                (coerce src-paths 'simple-vector)))
-           v))
-        (emit (asmstream-data-section asmstream) label `(.skip ,mark-index))))))
