@@ -132,22 +132,12 @@
   (declare (ignore env sexpr) #.+handler-optimize+)
   x)
 
-;;; Return a handler that returns a constant.
-;;; The %SEXPR constructor elides a handler for constants,
-;;; but there are cases where NIL sneaks through and demands a callable handler.
-;;; We avoid generating N copies of such handler. Same goes for 0 and T.
-(defun return-constant (object)
-  (cond ((null object) (load-time-value (handler #'%const nil)))
-        ((eq object t) (load-time-value (handler #'%const t)))
-        ((eql object 0) (load-time-value (handler #'%const 0)))
-        (t (handler #'%const object))))
-
 (defun return-constant-values (values env sexpr)
   (declare (ignore env sexpr) #.+handler-optimize+)
   (values-list values))
 
 ;; Forward defs
-(declaim (ftype function digest-form %eval eval-setq))
+(declaim (ftype function digest-form %eval eval-setq reutrn-constant))
 
 ;;; Return a SEXPR object that will, when dispatched, evaluate FORM.
 ;;; Unlike optimizations on SEXPRS whereby they install the most-specific
@@ -398,7 +388,7 @@
 ;;; forms, and an indicator of whether it is OK to create the block and bind
 ;;; variables at the same time. If not a block, return 0 and forms unaltered.
 ;;;
-;;; The optimization of eliding an extra ENV is performed only if no nontrival
+;;; The optimization of eliding an extra ENV is performed only if no nontrivial
 ;;; defaulting SEXPRs exist. A "trivial" one is either a constant or a reference
 ;;; to a variable (not a macro). Absent such a thing, nobody could notice that
 ;;; there aren't actually two different lexical contours, so creating variable
@@ -968,7 +958,7 @@ Test case.
 ;;; FNAME is the symbol at the head of the original form.
 ;;; Each element of KEYS is (#<FUNCTION> . SYM) or (SYM . SYMBOL-EXPANSION)
 ;;; The representation is unambiguous because a symbol is not a function,
-;;; whereas (SYM . EXPANSION|FUNCTION) is ambigious because through contortions
+;;; whereas (SYM . EXPANSION|FUNCTION) is ambiguous because through contortions
 ;;; it is possible to have a symbol's expansion be a function object.
 ;;; If any key is changed, restart using the original sexpr form.
 ;;;
@@ -1180,12 +1170,13 @@ Test case.
             (setf (debug-source-namestring source)
                   (sb-kernel::function-file-namestring f)
                   (debug-source-created source) nil) ; = unknown
-            ;; Clobber the TLF-NUM + OFFSET. We should do better that that,
-            ;; but interpreted functions do not track their file offset,
-            ;; so the conservative thing is to plug in NILs, not bogus values
-            ;; from whatever file (if any) is currently being loaded.
-            (setf (sb-c::compiled-debug-info-tlf-num+offset cdi) 0)
-            (setf (symbol-function fname) compiled-fun)))))
+            ;; Skirt a package lock by avoiding (SETF SYMBOL-FUNCTION)
+            ;; *technically* this should be a compare-and-swap to ensure that the
+            ;; original lambda expression is as expected. To do that, we need
+            ;; to get to the point where fdefns do not store "both" representations
+            ;; of one function pointer, because we can't assume that multi-word CAS
+            ;; is a thing.  Or use a mutex (horrible).
+            (setf (fdefn-fun (find-fdefn fname)) compiled-fun)))))
 
     (when (fluid-def-p fname)
       ;; Return a handler that calls FNAME very carefully
