@@ -127,12 +127,11 @@
 ;;; Type is either CTOR, for MAKE-INSTANCE, or ALLOCATOR, for ALLOCATE-INSTANCE
 (!defstruct-with-alternate-metaclass ctor
   :slot-names (type class-or-name class initargs state safe-p)
-  :boa-constructor %make-ctor
+  :constructor %make-ctor
   :superclass-name function
   :metaclass-name static-classoid
   :metaclass-constructor make-static-classoid
-  :dd-type funcallable-structure
-  :runtime-type-checks-p nil)
+  :dd-type funcallable-structure)
 
 ;;; All defined ctors.
 (defglobal *all-ctors* (make-hash-table :test #'equal
@@ -144,11 +143,11 @@
 
 ;;; Reset CTOR to use a default function that will compute an
 ;;; optimized constructor function when called.
-(defun install-initial-constructor (ctor &key force-p)
+(defun install-initial-constructor (ctor &optional force-p)
   (when (or force-p (ctor-class ctor))
     (setf (ctor-class ctor) nil
           (ctor-state ctor) 'initial)
-    (setf (funcallable-instance-fun ctor)
+    (setf (%funcallable-instance-fun ctor)
           (ecase (ctor-type ctor)
             (ctor
              (lambda (&rest args)
@@ -189,7 +188,7 @@
 ;;; Keep this a separate function for testing.
 (defun make-ctor (function-name class-name initargs safe-p)
   (let ((ctor (%make-ctor 'ctor class-name nil initargs nil safe-p)))
-    (install-initial-constructor ctor :force-p t)
+    (install-initial-constructor ctor t)
     (setf (gethash function-name *all-ctors*) ctor)
     ctor))
 
@@ -200,7 +199,7 @@
 
 (defun make-allocator (function-name class-name)
   (let ((ctor (%make-ctor 'allocator class-name nil nil nil nil)))
-    (install-initial-constructor ctor :force-p t)
+    (install-initial-constructor ctor t)
     (setf (gethash function-name *all-ctors*) ctor)
     ctor))
 
@@ -238,11 +237,12 @@
 (declaim (inline sxhash-symbol-or-class))
 (defun sxhash-symbol-or-class (x)
   (cond ((symbolp x) (sxhash x))
-        ((std-instance-p x) (sb-impl::std-instance-hash x))
-        ((fsc-instance-p x) (sb-impl::fsc-instance-hash x))
+        ((std-instance-p x) (sb-impl::instance-sxhash x))
+        ((fsc-instance-p x) (fsc-instance-hash x))
         (t
          (bug "Something strange where symbol or class expected."))))
 
+(export '(+ctor-list-max-size+ +ctor-table-max-size+)) ; for a test
 ;;; Max number of CTORs kept in an inline list cache. Once this is
 ;;; exceeded we switch to a table.
 (defconstant +ctor-list-max-size+ 12)
@@ -529,8 +529,8 @@
 ;;; SHARED-INITIALIZE methods.  One cannot initialize these variables
 ;;; to the right values here because said functions don't exist yet
 ;;; when this file is first loaded.
-(defvar *the-system-ii-method* nil)
-(defvar *the-system-si-method* nil)
+(define-load-time-global *the-system-ii-method* nil)
+(define-load-time-global *the-system-si-method* nil)
 
 (defun install-optimized-constructor (ctor)
   (with-world-lock ()
@@ -544,20 +544,17 @@
       ;; ...), because part of the deal is that those only happen from
       ;; FORCE-CACHE-FLUSHES, which create a new valid wrapper for the
       ;; class.  An invalid layout of T needs to be flushed, however.
-      (when (eq (layout-invalid (class-wrapper class)) t)
+      (when (eq (wrapper-invalid (class-wrapper class)) t)
         (%force-cache-flushes class))
       (setf (ctor-class ctor) class)
       (pushnew (make-weak-pointer ctor) (plist-value class 'ctors)
                :test #'eq :key #'weak-pointer-value)
       (multiple-value-bind (form locations names optimizedp)
           (constructor-function-form ctor)
-        (setf (funcallable-instance-fun ctor)
-              (apply
-               (let ((*compiling-optimized-constructor* t))
-                 (handler-bind ((compiler-note #'muffle-warning))
-                   (compile nil `(lambda ,names (declare #.*optimize-speed*)
-                                   ,form))))
-               locations)
+        (setf (%funcallable-instance-fun ctor)
+              (apply (let ((*compiling-optimized-constructor* t))
+                       (pcl-compile `(lambda ,names ,form) t))
+                     locations)
               (ctor-state ctor) (if optimizedp 'optimized 'fallback))))))
 
 (defun install-optimized-allocator (ctor)
@@ -572,17 +569,15 @@
       ;; ...), because part of the deal is that those only happen from
       ;; FORCE-CACHE-FLUSHES, which create a new valid wrapper for the
       ;; class.  An invalid layout of T needs to be flushed, however.
-      (when (eq (layout-invalid (class-wrapper class)) t)
+      (when (eq (wrapper-invalid (class-wrapper class)) t)
         (%force-cache-flushes class))
       (setf (ctor-class ctor) class)
       (pushnew (make-weak-pointer ctor) (plist-value class 'ctors)
                :test #'eq :key #'weak-pointer-value)
       (multiple-value-bind (form optimizedp)
           (allocator-function-form ctor)
-        (setf (funcallable-instance-fun ctor)
-              (let ((*compiling-optimized-constructor* t))
-                (handler-bind ((compiler-note #'muffle-warning))
-                  (compile nil form)))
+        (setf (%funcallable-instance-fun ctor)
+              (let ((*compiling-optimized-constructor* t)) (pcl-compile form t))
               (ctor-state ctor) (if optimizedp 'optimized 'fallback))))))
 
 (defun allocator-function-form (ctor)
@@ -754,22 +749,21 @@
        `(lambda ,(make-ctor-parameter-list ctor)
          (declare #.*optimize-speed*)
          (block nil
-           (when (layout-invalid ,wrapper)
-             (install-initial-constructor ,ctor)
+           (when (wrapper-invalid ,wrapper)
+             (install-initial-constructor ,ctor t)
              (return (funcall ,ctor ,@(make-ctor-parameter-list ctor))))
            ,(wrap-in-allocate-forms ctor body early-unbound-markers-p)))
        locations
        names
        t))))
 
-(defun optimizing-allocator-generator
-    (ctor)
+(defun optimizing-allocator-generator (ctor)
   (let ((wrapper (class-wrapper (ctor-class ctor))))
     `(lambda ()
        (declare #.*optimize-speed*)
        (block nil
-         (when (layout-invalid ,wrapper)
-           (install-initial-constructor ,ctor)
+         (when (wrapper-invalid ,wrapper)
+           (install-initial-constructor ,ctor t)
            (return (funcall ,ctor)))
          ,(wrap-in-allocate-forms ctor nil t)))))
 
@@ -780,24 +774,23 @@
 ;;; the instance, and .SLOTS. to the instance's slot vector around BODY.
 (defun wrap-in-allocate-forms (ctor body early-unbound-markers-p)
   (let* ((class (ctor-class ctor))
-         (wrapper (class-wrapper class))
-         (allocation-function (raw-instance-allocator class))
-         (slots-fetcher (slots-fetcher class)))
-    (if (eq allocation-function 'allocate-standard-instance)
-        `(let ((.instance. (%make-standard-instance nil #-compact-instance-header 0))
+         (wrapper (class-wrapper class)))
+    (etypecase class
+      (standard-class
+        `(let ((.instance. (%new-instance ,wrapper (1+ sb-vm:instance-data-start)))
                (.slots. (make-array
-                         ,(layout-length wrapper)
+                         ,(wrapper-length wrapper)
                          ,@(when early-unbound-markers-p
                                  '(:initial-element +slot-unbound+)))))
-           (setf (%instance-layout .instance.) ,wrapper)
-           (setf (std-instance-slots .instance.) .slots.)
+           (%instance-set .instance. sb-vm:instance-data-start .slots.)
            ,body
-           .instance.)
-        `(let* ((.instance. (,allocation-function ,wrapper))
-                (.slots. (,slots-fetcher .instance.)))
-           (declare (ignorable .slots.))
-           ,body
-           .instance.))))
+           .instance.))
+      (funcallable-standard-class
+        `(let* ((.instance. (allocate-standard-funcallable-instance ,wrapper nil))
+                (.slots. (fsc-instance-slots .instance.)))
+             (declare (ignorable .slots.))
+             ,body
+             .instance.)))))
 
 ;;; Return a form for invoking METHOD with arguments from ARGS.  As
 ;;; can be seen in METHOD-FUNCTION-FROM-FAST-FUNCTION, method
@@ -915,7 +908,7 @@
          (safe-p (ctor-safe-p ctor))
          (wrapper (class-wrapper class))
          (slot-vector
-          (make-array (layout-length wrapper) :initial-element nil))
+          (make-array (wrapper-length wrapper) :initial-element nil))
          (class-inits ())
          (default-inits ())
          (defaulting-initargs ())
@@ -1011,8 +1004,7 @@
                                (if (member slotd sbuc-slots :test #'eq)
                                    `(not (slot-boundp-using-class
                                           ,class .instance. ,slotd))
-                                   `(eq (clos-slots-ref .slots. ,i)
-                                        +slot-unbound+))))
+                                   `(unbound-marker-p (clos-slots-ref .slots. ,i)))))
                         (ecase kind
                           ((nil)
                            (unless early-unbound-markers-p
