@@ -211,7 +211,7 @@
 
 (defgeneric generic-function-lambda-list (generic-function))
 
-(defgeneric generic-function-pretty-arglist (generic-function))
+(defgeneric generic-function-pretty-arglist (generic-function &optional current-defmethod))
 
 (defgeneric gf-fast-method-function-p (gf))
 
@@ -227,12 +227,8 @@
 
 (defgeneric method-specializers (m))
 
-(defgeneric raw-instance-allocator (class))
-
 (defgeneric slot-definition-allocation (slotd))
 
-(defgeneric slots-fetcher (class))
-
 (defgeneric specializer-class (specializer))
 
 (defgeneric specializer-direct-generic-functions (specializer))
@@ -244,8 +240,6 @@
 (defgeneric specializer-method-holder (specializer &optional create))
 
 (defgeneric update-constructors (class))
-
-(defgeneric wrapper-fetcher (class))
 
 ;;;; 2 arguments
 
@@ -284,7 +278,27 @@
 
 (defgeneric map-dependents (metaobject function))
 
-(defgeneric parse-specializer-using-class (generic-function specializer-name))
+(defgeneric parse-specializer-using-class (generic-function specializer-name)
+  (:documentation
+   "Parse SPECIALIZER-NAME into a specializer object suitable for GENERIC-FUNCTION.
+
+If SPECIALIZER-NAME is not well-formed with respect to the specializer
+syntax of GENERIC-FUNCTION, an error of type
+SB-PCL:SPECIALIZER-NAME-SYNTAX-ERROR is signaled.
+
+If GENERIC-FUNCTION is a STANDARD-GENERIC-FUNCTION and
+SPECIALIZER-NAME is a symbol that does not name an existing class, an
+error of type SB-PCL:CLASS-NOT-FOUND-ERROR is signaled.
+
+Other errors may be signaled for generic function classes other than
+STANDARD-GENERIC-FUNCTION.
+
+If GENERIC-FUNCTION is a STANDARD-GENERIC-FUNCTION and
+SPECIALIZER-NAME is of the form (eql OBJECT), OBJECT is not
+evaluated (in contrast to DEFMETHOD's behavior).
+
+NOTE: This generic function is part of an SBCL-specific experimental
+protocol. Interface subject to change."))
 
 (defgeneric remove-boundp-method (class generic-function))
 
@@ -354,7 +368,6 @@
 
 (defgeneric specializer-type-specifier
     (proto-generic-function proto-method specializer)
-  #+sb-doc
   (:documentation
    "Return a type specifier for SPECIALIZER, a non-parsed specializer
 form or a SPECIALIZER instance.
