@@ -27,15 +27,6 @@
 ;;;; warranty about the software, its performance or its conformity to any
 ;;;; specification.
 
-(defpackage "SB-WALKER"
-  (:use "CL" "SB-INT" "SB-EXT")
-  (:documentation "internal: a code walker used by PCL")
-  (:export "DEFINE-WALKER-TEMPLATE"
-           "WALK-FORM"
-           "*WALK-FORM-EXPAND-MACROS-P*"
-           "VAR-LEXICAL-P" "VAR-SPECIAL-P"
-           "VAR-GLOBALLY-SPECIAL-P"
-           "VAR-DECLARATION"))
 (in-package "SB-WALKER")
 
 ;;;; forward references
@@ -89,7 +80,7 @@
 ;;; In SBCL, as in CMU CL before it, the environment is represented
 ;;; with a structure that holds alists for the functional things,
 ;;; variables, blocks, etc. Except for SYMBOL-MACROLET, only the
-;;; SB!C::LEXENV-FUNS slot is relevant. It holds: Alist (Name . What),
+;;; SB-C::LEXENV-FUNS slot is relevant. It holds: Alist (Name . What),
 ;;; where What is either a functional (a local function) or a list
 ;;; (MACRO . <function>) (a local macro, with the specifier expander.)
 ;;; Note that Name may be a (SETF <name>) function. Accessors are
@@ -97,7 +88,7 @@
 ;;;
 ;;; If WITH-AUGMENTED-ENVIRONMENT is called from WALKER-ENVIRONMENT-BIND
 ;;; this code hides the WALKER version of an environment
-;;; inside the SB!C::LEXENV structure.
+;;; inside the SB-C::LEXENV structure.
 ;;;
 ;;; In CMUCL (and former SBCL), This used to be a list of lists of form
 ;;; (<gensym-name> MACRO . #<interpreted-function>) in the :functions slot
@@ -233,9 +224,10 @@
 
 (defun convert-macro-to-lambda (llist body env &optional (name "dummy macro"))
   (declare (ignorable llist body env name))
-  #+sb-xc-host (error "CONVERT-MACRO-TO-LAMBDA called") ; no EVAL-IN-LEXENV
-  #-sb-xc-host
   (let ((gensym (make-symbol name)))
+    #+sb-xc-host
+    (eval `(sb-xc:defmacro ,gensym ,llist ,@body))
+    #-sb-xc-host
     (eval-in-lexenv `(defmacro ,gensym ,llist ,@body)
                     (sb-c::make-restricted-lexenv env))
     (macro-function gensym)))
@@ -301,21 +293,19 @@
   ;; if we're finding something in the real lexenv, we don't have a
   ;; bound declaration and so we specifically don't want to return
   ;; a special object that declarations can attach to, just the name.
-  (and env (find var (mapcar #'car (sb-c::lexenv-vars env)))))
+  (and env (find var (sb-c::lexenv-vars env) :key #'car)))
 
 (defun variable-symbol-macro-p (var env)
   ;; FIXME: crufty return convention
+  (declare (symbol var))
   (let ((entry (or (member var (env-lexical-variables env) :key #'car :test #'eq)
                    (and env (member var (sb-c::lexenv-vars env) :key #'car :test #'eq)))))
     (when (and (consp (cdar entry)) (eq (cadar entry) 'sb-sys:macro))
       (return-from variable-symbol-macro-p entry))
     (unless entry
-      (when (var-globally-symbol-macro-p var)
+      (when (eq (info :variable :kind var) :macro)
         (list (list* var 'sb-sys:macro (info :variable :macro-expansion var)))))))
 
-(defun var-globally-symbol-macro-p (var)
-  (eq (info :variable :kind var) :macro))
-
 (eval-when (:compile-toplevel :load-toplevel :execute)
 (defun walked-var-declaration-p (declaration)
   (member declaration '(sb-pcl::%class sb-pcl::%parameter
@@ -413,16 +403,13 @@
 ;;;       interface to its walker which looks like the interface to this
 ;;;       walker.
 
-(defmacro get-walker-template-internal (x)
-  `(get ,x 'walker-template))
-
 (defmacro define-walker-template (name
                                   &optional (template '(nil repeat (eval))))
-  `(setf (get-walker-template-internal ',name) ',template))
+  `(setf (info :function :walker-template ',name) ',template))
 
 (defun get-walker-template (x context)
   (cond ((symbolp x)
-         (get-walker-template-internal x))
+         (info :function :walker-template x))
         ((and (listp x) (eq (car x) 'lambda))
          '(lambda repeat (eval)))
         (t
@@ -454,8 +441,6 @@
 (define-walker-template macrolet             walk-macrolet)
 (define-walker-template multiple-value-call  (nil eval repeat (eval)))
 (define-walker-template multiple-value-prog1 (nil return repeat (eval)))
-(define-walker-template multiple-value-setq  walk-multiple-value-setq)
-(define-walker-template multiple-value-bind  walk-multiple-value-bind)
 (define-walker-template progn                (nil repeat (eval)))
 (define-walker-template progv                (nil eval eval repeat (eval)))
 (define-walker-template quote                (nil quote))
@@ -466,15 +451,123 @@
 (define-walker-template the                  (nil quote eval))
 (define-walker-template throw                (nil eval eval))
 (define-walker-template unwind-protect       (nil return repeat (eval)))
+(define-walker-template defun                walk-defun)
 
 ;;; SBCL-only special forms
 (define-walker-template truly-the (nil quote eval))
 (define-walker-template sb-kernel:the* (nil quote eval))
+(define-walker-template sb-c::with-source-form (nil nil eval))
 ;;; FIXME: maybe we don't need this one any more, given that
 ;;; NAMED-LAMBDA now expands into (FUNCTION (NAMED-LAMBDA ...))?
 (define-walker-template named-lambda walk-named-lambda)
+#|
+;;; To find templateized symbols that aren't special operators:
+(do-all-symbols (s)
+  (let ((template
+         (sb-int:info :function :walker-template s)))
+    (when (and template (not (special-operator-p s)))
+      (format t "Why? ~S~%" s))))
+|#
 
 (defvar *walk-form-expand-macros-p* nil)
+(defvar *walk-form-preserve-source* nil)
+
+(defun macroexpand-all (form &optional environment)
+  (let ((*walk-form-expand-macros-p* t))
+    (sb-walker:walk-form
+     form environment
+     (lambda (subform context env)
+       (acond ((and (eq context :eval)
+                    (listp subform)
+                    (symbolp (car subform))
+                    (get (car subform) :partial-macroexpander))
+               ;; The partial expander must return T as its second value
+               ;; if it wants to stop the walk.
+               (funcall it subform env))
+              (t
+               subform))))))
+
+;; Given EXPR, the argument to an invocation of Quasiquote macro, macroexpand
+;; evaluable subforms of EXPR using ENV. A subform is evaluable if all
+;; preceding occurrences of #\` have been "canceled" by a comma.
+;; DEPTH counts the nesting and should not be supplied by external callers.
+(defun %quasiquoted-macroexpand-all (expr env &optional (depth 0))
+  (flet ((quasiquote-p (x)
+           (and (listp x) (eq (car x) 'quasiquote) (singleton-p (cdr x))))
+         (recurse (x)
+           (%quasiquoted-macroexpand-all x env depth)))
+    (if (atom expr)
+        (cond ((simple-vector-p expr) (map 'vector #'recurse expr))
+              ((comma-p expr)
+               (unquote (if (> depth 1)
+                            (%quasiquoted-macroexpand-all
+                             (comma-expr expr) env (1- depth))
+                            (macroexpand-all (comma-expr expr) env))
+                        (comma-kind expr)))
+              (t expr))
+        (if (quasiquote-p expr)
+            (list 'quasiquote
+                  (%quasiquoted-macroexpand-all (second expr) env (1+ depth)))
+            (let (result)
+              (loop
+               (push (recurse (pop expr)) result)
+               (when (or (atom expr) (quasiquote-p expr))
+                 (return (nreconc result (recurse expr))))))))))
+
+(setf (get 'quasiquote :partial-macroexpander)
+      (lambda (form env)
+        (destructuring-bind (arg) (cdr form) ; sanity-check the shape
+          (declare (ignore arg))
+          (values (%quasiquoted-macroexpand-all form env) t))))
+
+#|
+
+;; Another example that some people might find useful.
+
+(defun macroexpand-decls+forms (body env) ; a bit of a kludge, but it works
+  (mapcar (lambda (x)
+            (if (and (listp x) (eq (car x) 'declare))
+                x
+                (macroexpand-all x env)))
+          body))
+
+(setf (get 'dotimes :partial-macroexpander)
+      (lambda (form env)
+        (destructuring-bind ((var count &optional (result nil result-p))
+                             &body body) (cdr form)
+            (values `(dotimes (,var ,(macroexpand-all count env)
+                               ,@(if result-p
+                                     (list (macroexpand-all result env))))
+                       ,@(macroexpand-decls+forms body env))
+                    t))))
+
+(macroexpand-all '(macrolet ((hair (x) `(car ,x)))
+                   (dotimes (i (bar)) (foo i (hair baz)) l))))
+=>
+(MACROLET ((HAIR (X)
+             `(CAR ,X)))
+  (DOTIMES (I (BAR)) (FOO I (CAR BAZ)) L))
+
+instead of
+
+(MACROLET ((HAIR (X)
+             `(CAR ,X)))
+  (BLOCK NIL
+    (LET ((I 0) (#:COUNT699 (BAR)))
+      (DECLARE (TYPE UNSIGNED-BYTE I)
+               (TYPE INTEGER #:COUNT699))
+      (TAGBODY
+        (GO #:G701)
+       #:G700
+        (TAGBODY (FOO I (CAR BAZ)) L)
+        (LET* ()
+          (MULTIPLE-VALUE-BIND (#:NEW702) (1+ I) (PROGN (SETQ I #:NEW702) NIL)))
+       #:G701
+        (IF (>= I #:COUNT699)
+            NIL
+            (PROGN (GO #:G700)))
+        (RETURN-FROM NIL (PROGN NIL))))))
+|#
 
 #+sb-fasteval
 (declaim (ftype (sfunction (sb-interpreter:basic-env &optional t) sb-kernel:lexenv)
@@ -512,63 +605,83 @@
   ;; by walk-function is T then we don't recurse...
   (catch form
     (with-current-source-form (form)
-      (multiple-value-bind (newform walk-no-more-p)
-          (funcall (env-walk-function env) form context env)
-        (catch newform
-          (cond
-            (walk-no-more-p newform)
-            ((not (eq form newform))
-             (walk-form-internal newform context env))
-            ((and (not (consp newform))
-                  (or (eql context :eval)
-                      (eql context :set)))
-             (let ((symmac (car (variable-symbol-macro-p newform env))))
-               (if symmac
-                   (let* ((newnewform (walk-form-internal (cddr symmac)
-                                                          context
-                                                          env))
-                          (resultform
-                           (if (eq newnewform (cddr symmac))
-                               (if *walk-form-expand-macros-p* newnewform newform)
-                               newnewform))
-                          (type (env-var-type newform env)))
-                     (if (eq t type)
-                         resultform
-                         `(the ,type ,resultform)))
-                   newform)))
-            ((eql context :set) newform)
-            (t
-             (let* ((fn (car newform))
-                    (template (get-walker-template fn newform)))
-               (if template
-                   (if (symbolp template)
-                       (funcall template newform context env)
-                       (walk-template newform template context env))
-                   (multiple-value-bind (newnewform macrop)
-                       (walker-environment-bind
-                           (new-env env :walk-form newform)
-                         (%macroexpand-1 newform new-env))
-                     (cond
-                       (macrop
-                        (let ((newnewnewform (walk-form-internal newnewform
-                                                                 context
-                                                                 env)))
-                          (if (eq newnewnewform newnewform)
-                              (if *walk-form-expand-macros-p* newnewform newform)
-                              newnewnewform)))
-                       ((and (symbolp fn)
-                             (special-operator-p fn))
-                        ;; This shouldn't happen, since this walker is now
-                        ;; maintained as part of SBCL, so it should know
-                        ;; about all the special forms that SBCL knows
-                        ;; about.
-                        (bug "unexpected special form ~S" fn))
-                       (t
-                        ;; Otherwise, walk the form as if it's just a
-                        ;; standard function call using a template for
-                        ;; standard function call.
-                        (walk-template
-                         newnewform '(call repeat (eval)) context env)))))))))))))
+      (let ((sb-c::*compiler-error-bailout*
+              (lambda (c)
+                (return-from walk-form-internal
+                  (sb-c::make-compiler-error-form c form)))))
+        (multiple-value-bind (newform walk-no-more-p)
+            (funcall (env-walk-function env) form context env)
+          (catch newform
+            (cond
+              (walk-no-more-p newform)
+              ((not (eq form newform))
+               (record-new-source-path form newform)
+               (walk-form-internal newform context env))
+              ((and (not (consp newform))
+                    (or (eql context :eval)
+                        (eql context :set)))
+               (let ((symmac (and (symbolp newform)
+                                  (car (variable-symbol-macro-p newform env)))))
+                 (if symmac
+                     (let* ((newnewform (walk-form-internal (cddr symmac)
+                                                            context
+                                                            env))
+                            (resultform
+                              (if (eq newnewform (cddr symmac))
+                                  (if *walk-form-expand-macros-p* newnewform newform)
+                                  newnewform))
+                            (type (env-var-type newform env)))
+                       (if (eq t type)
+                           resultform
+                           `(the ,type ,resultform)))
+                     newform)))
+              ((eql context :set) newform)
+              (t
+               (let* ((fn (car newform))
+                      (template (get-walker-template fn newform)))
+                 (if template
+                     (if (symbolp template)
+                         (funcall template newform context env)
+                         (walk-template newform template context env))
+                     (multiple-value-bind (newnewform macrop)
+                         (walker-environment-bind
+                             (new-env env :walk-form newform)
+                           (%macroexpand-1 newform new-env))
+                       (cond
+                         (macrop
+                          (let ((newnewnewform (walk-form-internal newnewform
+                                                                   context
+                                                                   env)))
+                            (cond ((eq newnewnewform newnewform)
+                                   (if *walk-form-expand-macros-p* newnewform newform))
+                                  (t
+                                   (record-new-source-path newform newnewnewform )))))
+                         ((and (symbolp fn)
+                               (special-operator-p fn))
+                          ;; This shouldn't happen, since this walker is now
+                          ;; maintained as part of SBCL, so it should know
+                          ;; about all the special forms that SBCL knows
+                          ;; about.
+                          (bug "unexpected special form ~S" fn))
+                         (t
+                          ;; Otherwise, walk the form as if it's just a
+                          ;; standard function call using a template for
+                          ;; standard function call.
+                          (walk-template
+                           newnewform '(call repeat (eval)) context env))))))))))))))
+
+(defun record-new-source-path (old-form new-form)
+  (when (and *walk-form-preserve-source*
+             (boundp 'sb-c::*source-paths*))
+    (let ((path (gethash old-form sb-c::*source-paths*)))
+      (when path
+        (setf (gethash new-form sb-c::*source-paths*) path))))
+  new-form)
+
+(defun recons (old-cons car cdr)
+  (if (and (eq car (car old-cons)) (eq cdr (cdr old-cons)))
+      old-cons
+      (record-new-source-path old-cons (cons car cdr))))
 
 (defun walk-template (form template context env)
   (if (atom template)
@@ -758,23 +871,15 @@
               new-env)))
       (relist* form let walked-bindings walked-body))))
 
-(defun let*-binding-name (binding)
-  (if (symbolp binding)
-      binding
-      (car binding)))
-
-(defun let*-binding-init (binding)
-  (if (or (symbolp binding)
-          (null (cdr binding)))
-      'no-init
-      (cadr binding)))
-
-(defun let*-bindings (bindings &aux names inits (seen (make-hash-table :test #'eq)))
+(defun let*-bindings (bindings &aux names inits (seen (alloc-xset)))
   (dolist (binding (reverse bindings) (values names inits))
-    (let ((name (let*-binding-name binding)))
-      (push (cons name (gethash name seen)) names)
-      (setf (gethash name seen) t)
-      (push (let*-binding-init binding) inits))))
+    (multiple-value-bind (name init)
+        (cond ((atom binding) (values binding 'no-init))
+              ((not (cdr binding)) (values (car binding) 'no-init))
+              (t (values (car binding) (cadr binding))))
+      (push (cons name (xset-member-p name seen)) names)
+      (add-to-xset name seen)
+      (push init inits))))
 
 (defun walk-let* (form context env)
   (walker-environment-bind (new-env env)
@@ -851,55 +956,24 @@
       (relist*
        form locally walked-body))))
 
-(defun walk-multiple-value-setq (form context env)
-  (let ((vars (cadr form)))
-    (if (some (lambda (var)
-                (variable-symbol-macro-p var env))
-              vars)
-        (let* ((temps (mapcar (lambda (var)
-                                (declare (ignore var))
-                                (gensym))
-                              vars))
-               (sets (mapcar (lambda (var temp) `(setq ,var ,temp))
-                             vars
-                             temps))
-               (expanded `(multiple-value-bind ,temps ,(caddr form)
-                             ,@sets))
-               (walked (walk-form-internal expanded context env)))
-          (if (eq walked expanded)
-              form
-              walked))
-        (walk-template form '(nil (repeat (set)) eval) context env))))
-
-(defun walk-multiple-value-bind (form context old-env)
-  (walker-environment-bind (new-env old-env)
-    (let* ((mvb (car form))
-           (bindings (cadr form))
-           (mv-form (walk-template (caddr form) 'eval context old-env))
-           (body (cdddr form))
-           walked-bindings
-           (walked-body
-             (walk-declarations
-               body
-               (lambda (real-body real-env)
-                 (setq walked-bindings
-                       (walk-bindings-1 bindings old-env new-env context))
-                 (walk-repeat-eval real-body real-env))
-               new-env)))
-      (relist* form mvb walked-bindings mv-form walked-body))))
-
 (defun walk-bindings-1 (bindings old-env new-env context)
   (and bindings
        (let ((binding (car bindings)))
          (recons bindings
-                 (if (symbolp binding)
+                 (typecase binding
+                   (symbol
                      (prog1 binding
-                       (note-var-binding binding new-env))
-                     (prog1 (relist binding
-                                    (car binding)
-                                    (walk-form-internal
-                                     (cadr binding) context old-env))
+                       (note-var-binding binding new-env)))
+                   ((cons symbol list)
+                     (prog1 (relist* binding
+                                     (car binding)
+                                     (walk-form-internal
+                                      (cadr binding) context old-env)
+                                     ;; Preserve "trailing junk"
+                                     (cddr binding))
                        (note-var-binding (car binding) new-env)))
+                   (t ; illegal syntax, (let (#(foo))) or (let (a . #*1)) etc
+                    binding))
                  (walk-bindings-1 (cdr bindings) old-env new-env context)))))
 
 (defun walk-lambda (form context old-env)
@@ -928,6 +1002,10 @@
                walked-arglist
                walked-body))))
 
+;;; The DEFUN macro does some stuff with the environment, handle it here.
+(defun walk-defun (form context env)
+  (recons form (car form) (walk-lambda (cdr form) context env)))
+
 (defun walk-setq (form context env)
   (if (cdddr form)
       (let* ((expanded (let ((rforms nil)
@@ -941,7 +1019,7 @@
             `(progn ,@walked)))
       (let* ((var (cadr form))
              (val (caddr form))
-             (symmac (car (variable-symbol-macro-p var env))))
+             (symmac (and (symbolp var) (car (variable-symbol-macro-p var env)))))
         (if symmac
             (let* ((type (env-var-type var env))
                    (expanded (if (eq t type)
