@@ -69,16 +69,15 @@ int arch_os_thread_init(struct thread *thread) {
     set_data_desc_addr(&ldt_entry, thread);
     set_data_desc_size(&ldt_entry, dynamic_values_bytes);
 
-    thread_mutex_lock(&modify_ldt_lock);
+    ignore_value(mutex_acquire(&modify_ldt_lock));
     n = i386_set_ldt(LDT_AUTO_ALLOC, (union ldt_entry*) &ldt_entry, 1);
 
     if (n < 0) {
         perror("i386_set_ldt");
         lose("unexpected i386_set_ldt(..) failure");
     }
-    thread_mutex_unlock(&modify_ldt_lock);
+    ignore_value(mutex_release(&modify_ldt_lock));
 
-    FSHOW_SIGNAL((stderr, "/ TLS: Allocated LDT %x\n", n));
     thread->tls_cookie=n;
     arch_os_load_ldt(thread);
 #endif
@@ -105,12 +104,10 @@ int arch_os_thread_cleanup(struct thread *thread) {
     /* Set the %%fs register back to 0 and free the ldt by setting it
      * to NULL.
      */
-    FSHOW_SIGNAL((stderr, "/ TLS: Freeing LDT %x\n", n));
-
     __asm__ __volatile__ ("mov %0, %%fs" : : "r"(0));
-    thread_mutex_lock(&modify_ldt_lock);
+    ignore_value(mutex_acquire(&modify_ldt_lock));
     i386_set_ldt(n, NULL, 1);
-    thread_mutex_unlock(&modify_ldt_lock);
+    ignore_value(mutex_release(&modify_ldt_lock));
 #endif
     return 1;                  /* success */
 }
